# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""VIIRS NetCDF reader.

This VIIRS reader is designed for reading the NPP/JPSS VIIRS files geoips.
The reader is only using the python functions and xarray variables.  Although
the file name indicates the data is in netcdf4 format.

Thus, the reader is based on the netcdf4 data format.

The orginal reader (viirs_aotcimss_ncdf4_reader.py) was developed for geoips1,
which applied many geoips1 function.

V1.0:  NRL-Monterey, 09/17/2020

VIIRS file infOrmation::

    There are 6 files for each time of VIIRS data, i.e.,
    For NASA NPP VIIRS
    20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02DNB.sdr.x.x.nc
    20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02IMG.sdr.x.x.nc
    20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02MOD.sdr.x.x.nc
    20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP03DNB.sdr.x.x.nc
    20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP03IMG.sdr.x.x.nc
    20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP03MOD.sdr.x.x.nc

    For JPSS VIIRS
    20200914.150000.npp.viirs.viirs_sips_jpss_uwssec_001.x.VJ102DNB.sdr.x.x.nc
    20200914.150000.npp.viirs.viirs_sips_jpss_uwssec_001.x.VJ102IMG.sdr.x.x.nc
    20200914.150000.npp.viirs.viirs_sips_jpss_uwssec_001.x.VJ102MOD.sdr.x.x.nc
    20200914.150000.npp.viirs.viirs_sips_jpss_uwssec_001.x.VJ103DNB.sdr.x.x.nc
    20200914.150000.npp.viirs.viirs_sips_jpss_uwssec_001.x.VJ103IMG.sdr.x.x.nc
    20200914.150000.npp.viirs.viirs_sips_jpss_uwssec_001.x.VJ103MOD.sdr.x.x.nc

    DNB: VIIRS day-night Band obs
    MOD: VIIRS M-Band  obs
    IMG: VIIRS I-Band  obs

The *.VNP02* files are for the data records, while the *.VNP03* files are for
the geolocation data records.

The xarray of geoips reader need both the data and lat/lon info. Thus, this
VIIRS reader is designed to read in the paired VNP02 and VNP03 files, depending
on any one of DNB or IMG or MOD file.  In order to minimize dupilcated excution
of VIIRS files, additional adjust of excution of the VIIRS files will be needed
(discussion with Mindy on how to do it).
"""
# Python Standard Libraries
from collections import defaultdict
from datetime import datetime
import logging
import os

# Third-Party Libraries
import numpy
import pandas as pd
import xarray as xr

# GeoIPS imports
from geoips.plugins.modules.readers.utils.geostationary_geolocation import get_indexes
from geoips.utils.context_managers import import_optional_dependencies
from geoips.errors import MissingRequiredXarrayMetadata

# If this reader is not installed on the system, don't fail altogether, just skip this
# import. This reader will not work if the import fails, and the package will have to be
# installed to process data of this type.
LOG = logging.getLogger(__name__)

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    import netCDF4 as ncdf


# @staticmethod                                     # not sure where it is uwas used?


VARLIST = {
    "DNB": ["DNB_observations"],
    "IMG": ["I04", "I05"],
    "IMG-Vis": ["I01", "I02", "I03"],
    "MOD": ["M07", "M08", "M10", "M11", "M12", "M13", "M14", "M15", "M16"],
    "MOD-Vis": ["M01", "M02", "M03", "M04", "M05", "M06", "M09"],
}

DNB_CHANNELS = ["DNB_observations"]
BT_CHANNELS = ["I04", "I05", "M12", "M13", "M14", "M15", "M16"]
REF_CHANNELS = [
    "I01",
    "I02",
    "I03",
    "M01",
    "M02",
    "M03",
    "M04",
    "M05",
    "M06",
    "M07",
    "M08",
    "M09",
    "M10",
    "M11",
]

# List of geolocation variables

GVARLIST = {
    "DNB": [
        "latitude",
        "longitude",
        "solar_zenith",
        "solar_azimuth",
        "sensor_zenith",
        "sensor_azimuth",
        "lunar_zenith",
        "moon_phase_angle",
    ],
    "IMG": [
        "latitude",
        "longitude",
        "solar_zenith",
        "solar_azimuth",
        "sensor_zenith",
        "sensor_azimuth",
    ],
    "IMG-Vis": [
        "latitude",
        "longitude",
        "solar_zenith",
        "solar_azimuth",
        "sensor_zenith",
        "sensor_azimuth",
    ],
    "MOD": [
        "latitude",
        "longitude",
        "solar_zenith",
        "solar_azimuth",
        "sensor_zenith",
        "sensor_azimuth",
    ],
    "MOD-Vis": [
        "latitude",
        "longitude",
        "solar_zenith",
        "solar_azimuth",
        "sensor_zenith",
        "sensor_azimuth",
    ],
}
xvarnames = {
    "solar_zenith": "solar_zenith_angle",
    "solar_azimuth": "solar_azimuth_angle",
    "sensor_zenith": "satellite_zenith_angle",
    "sensor_azimuth": "satellite_azimuth_angle",
    "lunar_zenith": "lunar_zenith_angle",
    "DNB_observations": "DNB",
}

# IMG:
#     I01(nline, npix): I-band 01 earth view reflectance (0-65527).
#                                                                 range of vlaues: 0 - 1
#                       add_offset=0,  scale_factor=1.9991758e-05,
#                       radiance_scale_factor=0.010167242, radiance_add_offset=0,
#                       radiance_units: Watts/meter^2/steradian/micrometer
#     I02(nline, npix): I-band 02 earth view reflectance (0-65527).
#                       add_offset=0,  scale_factor=1.9991758e-05,
#                       radiance_scale_factor=0.006007384, radiance_add_offset=0,
#                       radiance_units: Watts/meter^2/steradian/micrometer
#     I03(nline, npix): I-band 03 earth view reflectance (0-65527).
#                       add_offset=0,  scale_factor=1.9991758e-05,
#                       radiance_scale_factor=0.0015311801, radiance_add_offset=0,
#                       radiance_units: Watts/meter^2/steradian/micrometer
#     I04(nline, npix): I-band 04 earth view radiance (0-65527). need LUT for TB
#                                                                             conversion
#                       scale_factor=6.104354e-05, add_offset=0.0016703
#                       units: Watts/meter^2/steradian/micrometer
#                       I04_brightness_temperature_lut(65536). LUT[65535] is for
#                                                                          masked point.
#     I05(nline, npix): I-band 05 earth view radiance (0-65527). need LUT for TB
#                                                                             conversion
#                       scale_factor=0.0003815221, add_offset=0.141121
#                       units: Watts/meter^2/steradian/micrometer
#                       I05_brightness_temperature_lut(65536). LUT[65535] is for
#                                                                          masked point.
#     I04TB/I05TB     : associated TBs for I04/I05
#
#     Note: I04 and I05 converstion to TB will apply its TB conversion LUT. Since the
#                       change of TB is so small bewteen the LUT index, we just take
#                       value fo the I04/I05 as the LUT index (auto roundup) for easy
#                       process. This way will have an error of TB less than 0.05K.
#                       Otherwsie, a interpolation is needed for a more accurate TB.

# MOD:

#     M01 - M11(nline,npix):  M-band 01-11 earth view reflectance
#     M12 - M16(nline,npix):  M-band 12-16 earth view radiance. will convert radiance
#                                                                               anto TBs
#     M12TB-M16TB:         :  associated TBs for M12-16
#

#    Run plan:  fname has a list of VIIRS files (3 *02* or *03* files)

interface = "readers"
family = "standard"
name = "viirs_netcdf"
source_names = ["viirs"]


def _get_geolocation_metadata(orig_shape, fnames, xarray):
    """
    Gather all of the metadata used in creating geolocation data for input file.

    This is split out so we can easily create a chash of the data for creation
    of a unique filename. This allows us to avoid recalculation of angles that
    have already been calculated.
    """
    geomet = {}
    geomet["platform_name"] = "viirs"
    geomet["scene"] = "stitched_granules"
    # Just getting the nadir resolution in kilometers.  Must extract from a string.
    geomet["res_km"] = xarray.attrs["resolution_km"]
    geomet["roi_factor"] = 5  # roi = res * roi_factor, was 10
    geomet["num_lines"] = orig_shape[0]
    geomet["num_samples"] = orig_shape[1]
    geomet["fnames"] = fnames
    return geomet


def required_chan(chans, varnames):
    """Return True if required channel."""
    if chans is None:
        return True
    for varname in varnames:
        if varname in chans:
            return True
    return False


def required_geo_chan(xarrays, xvarname):
    """Return True if required geolocation channel."""
    # for data_type in list(xarrays.keys()):
    #     if xvarname in list(xarrays[data_type].keys()):
    #         LOG.info('        SKIPPING %s geolocation channel %s, xarray GEOLOCATION
    #                           variable %s exists', data_type, xvarname, xvarname)
    #         return False
    return True


def required_geo(chans, data_type):
    """Return True if required geolocation dataset."""
    if chans is None:
        return True
    for varname in chans:
        if (
            varname.replace("Ref", "").replace("Rad", "").replace("BT", "")
            in VARLIST[data_type]
        ):
            return True
        if (
            varname.replace("Ref", "").replace("Rad", "").replace("BT", "")
            + "_observations"
            in VARLIST[data_type]
        ):
            return True
    return False


def add_to_xarray(varname, nparr, xobj, dataset_masks, data_type, nparr_mask):
    """Add variable to xarray Dataset."""
    # Add current dataset to xarray object
    if varname not in xobj.variables:
        merged_array = nparr
    else:
        merged_array = numpy.vstack([xobj[varname].to_masked_array(), nparr])
    xobj[varname] = xr.DataArray(
        merged_array, dims=["dim_" + str(merged_array.shape[0]), "dim_1"]
    )

    # If nparr_mask.shape != nparr.shape, then it is "False" so create an
    # array of False the size of nparr
    if nparr_mask.shape != nparr.shape:
        nparr_mask = numpy.zeros(nparr.shape).astype(bool)
    # If we haven't merged nparr_mask into dataset_masks[data_type] yet, do it now
    if dataset_masks[data_type] is False:
        dataset_masks[data_type] = nparr_mask
    elif dataset_masks[data_type].shape != merged_array.shape:
        # This was previously only being applied to "BT_CHANNELS", which
        # caused the unprojected test scripts to fail (due to incorrect
        # shape of the final mask array).  Removing that check so ALL
        # variables are vstacked.
        dataset_masks[data_type] = numpy.vstack([dataset_masks[data_type], nparr_mask])


def call(
    fnames,
    metadata_only=False,
    chans=None,
    area_def=None,
    self_register=False,
    resample=False,
):
    """Read VIIRS netcdf data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # since fname is a LIST of input files, this reader needs additional adjustments to
    # read all files and put them into the XARRAY output (add one more array for
    # number of files)

    """
    fnames=['readers/data_viirs/20200916.081200.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02DNB.sdr.x.x.nc',\
           'readers/data_viirs/20200916.081200.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02IMG.sdr.x.x.nc',\
           'readers/data_viirs/20200916.081200.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02MOD.sdr.x.x.nc']

    fnames=['data_viirs/20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02DNB.sdr.x.x.nc',\
           'data_viirs/20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02IMG.sdr.x.x.nc',\
           'data_viirs/20200826.074800.npp.viirs.viirs_npp_nasaearthdata_x.x.VNP02MOD.sdr.x.x.nc']
    """

    # --------------- loop input files ---------------
    xarrays = {}
    dataset_masks = {}
    LOG.info("Requested Channels: %s", chans)

    tracked_data_type_vars = defaultdict(dict)
    # This fails if fnames happens to be in a different order
    for fname in sorted(fnames):
        # LOG.info('tst name= ', fname)

        # # chech for right VIIRS file
        # if 'viirs' in os.path.basename(fname):
        #     LOG.info('found a VIIRS file')
        # else:
        #     LOG.info('not a VIIRS file: skip it')
        #     raise

        # open the paired input files
        ncdf_file = ncdf.Dataset(str(fname), "r")

        LOG.info("    Trying file %s", fname)

        # *************** input VIRRS variables *****************

        # read IMG  or MOD  or DNB files

        # if data_type == 'IMG' or data_type == 'MOD' or data_type == 'DNB':

        # dict_var =dict.fromkeys(VARLIST[data_type])
        # dict_gvar=dict.fromkeys(GVARLIST[data_type])

        data_type = ncdf_file.product_name[5:8]

        if data_type not in dataset_masks:
            dataset_masks[data_type] = False
            xarrays[data_type] = xr.Dataset()
            xarrays[data_type].attrs["source_file_names"] = []

        #  ------  Time info for each pixel ------
        #  Start (end) time of VIIRS scan in seconds since 1/1/1993
        #  Python datetime64 time starts on '1970-01-01T00:00:00Z'
        #  Thus, 1/1/1993 datetime64 has seconds of 725846400
        #  The scan info does not match the line info of each fields (lines x pix)
        # thus, no tiemstamp is applied.  Just the Start_date and End_date are
        # sued for time info.

        scan_start_time = ncdf_file["scan_line_attributes"].variables[
            "scan_start_time"
        ][:]
        scan_end_time = ncdf_file["scan_line_attributes"].variables["scan_end_time"][:]

        # scan = ncdf_file.groups["scan_line_attributes"]
        # stime = scan.variables['scan_start_time'][:]
        # etime = scan.variables['scan_end_time'][:]

        s_time = scan_start_time + 725846400
        e_time = scan_end_time + 725846400

        # convert it to UTC time in format: datetime.datetime(2020, 8, 26, 7,
        #                                                     48, 10, 445420)
        Start_date = datetime.utcfromtimestamp(s_time[0])
        End_date = datetime.utcfromtimestamp(e_time[-1])

        #    ************  setup VIIRS xarray variables *****************
        #    only the avaialble fileds are arranged in the xaaray object
        #         the missing fields in the data file are not included.
        #         i.e, if I01-03 are missing the *IMG file, then only I04 and
        #              I05 fields will be processed and output to Xarray

        # from IPython import embed as shell

        xarrays[data_type].attrs["start_datetime"] = Start_date
        xarrays[data_type].attrs["end_datetime"] = End_date
        xarrays[data_type].attrs["source_name"] = "viirs"
        if ncdf_file.platform == "Suomi-NPP":
            xarrays[data_type].attrs["platform_name"] = "npp"
        if ncdf_file.platform == "JPSS-1":
            # xarrays[data_type].attrs['platform_name']  = 'jpss-1'
            # Attribute still lists JPSS-1, but operational satellite name is NOAA-20.
            xarrays[data_type].attrs["platform_name"] = "noaa-20"
        if ncdf_file.platform == "JPSS-2":
            # xarrays[data_type].attrs['platform_name']  = 'jpss-2'
            # Attribute still lists JPSS-2, but operational satellite name is NOAA-21.
            xarrays[data_type].attrs["platform_name"] = "noaa-21"
        if "platform_name" not in xarrays[data_type].attrs:
            raise MissingRequiredXarrayMetadata(
                "Missing platform_name, ensure platform_name set in reader"
            )
        xarrays[data_type].attrs["data_provider"] = "NASA"
        if os.path.basename(fname) not in xarrays[data_type].attrs["source_file_names"]:
            xarrays[data_type].attrs["source_file_names"] += [os.path.basename(fname)]

        # MTIFs need to be "prettier" for PMW products, so 2km resolution for
        # final image
        xarrays[data_type].attrs["sample_distance_km"] = 2
        xarrays[data_type].attrs["interpolation_radius_of_influence"] = 3000

        if metadata_only is True:
            LOG.info(
                "metadata_only is True, reading only first file for metadata "
                "information and returning"
            )
            return {"METADATA": xarrays[data_type]}

        # check fo avaialble varaibles
        if "observation_data" in ncdf_file.groups.keys():
            list_vars = ncdf_file["observation_data"].variables.keys()
            ncdata = ncdf_file["observation_data"]
        if "geolocation_data" in ncdf_file.groups.keys():
            list_vars = ncdf_file["geolocation_data"].variables.keys()
            ncdata = ncdf_file["geolocation_data"]

        # Assign visible chans to own data_type, since they will disappear
        # from day -> night.
        # These are located in the VNP02MOD/IMG files, or the MOD dataset type
        # Also keep track of all variables in the parent file of the data_type
        # This will help determine if we need to add geolocation vars to the
        # data_type xarray downstream
        default_data_type = data_type
        vis_data_type = data_type + "-Vis"
        if Start_date not in tracked_data_type_vars[default_data_type]:
            tracked_data_type_vars[default_data_type][Start_date] = []
        tracked_data_type_vars[default_data_type][Start_date].extend(list(list_vars))
        if data_type in ["MOD", "IMG"]:
            if vis_data_type not in dataset_masks:
                dataset_masks[vis_data_type] = False
                xarrays[vis_data_type] = xr.Dataset()
                xarrays[vis_data_type].attrs["source_file_names"] = []
            xarrays[vis_data_type].attrs = xarrays[data_type].attrs
            geo_target_data_types = [default_data_type, vis_data_type]
            vis_vars = VARLIST[vis_data_type]
            if Start_date not in tracked_data_type_vars[vis_data_type]:
                tracked_data_type_vars[vis_data_type][Start_date] = []
            tracked_data_type_vars[vis_data_type][Start_date].extend(list(list_vars))
        else:
            geo_target_data_types = [default_data_type]
            vis_vars = []

        for var in list_vars:
            xvarname = var
            if var in xvarnames:
                xvarname = xvarnames[var]

            btvarname = xvarname + "BT"
            refvarname = xvarname + "Ref"
            radvarname = xvarname + "Rad"

            ncvar = ncdata.variables[var]

            if var in vis_vars:
                data_type = vis_data_type
            else:
                data_type = default_data_type

            # Need Rad in order to calculate DNB Ref variable.
            if var in DNB_CHANNELS and required_chan(chans, [radvarname, refvarname]):
                LOG.info(
                    "        Reading %s channel %s into DNB variable %s",
                    data_type,
                    var,
                    radvarname,
                )
                nparr = numpy.ma.masked_greater(ncvar[...], ncvar.valid_max)
                add_to_xarray(
                    radvarname,
                    nparr,
                    xarrays[data_type],
                    dataset_masks,
                    data_type,
                    nparr.mask,
                )
                for attrname in ncvar.ncattrs():
                    xarrays[data_type][radvarname].attrs[attrname] = ncvar.getncattr(
                        attrname
                    )

            # Apply Brightness Temperature conversions if BT version of variable was
            # requested.
            if (
                var in BT_CHANNELS + REF_CHANNELS
                and required_chan(chans, [btvarname])
                and var + "_brightness_temperature_lut" in ncdata.variables.keys()
            ):
                LOG.info(
                    "        Reading %s channel %s into BRIGHTNESS TEMPERATURE "
                    "variable %s",
                    data_type,
                    var,
                    btvarname,
                )
                btlut = ncdata.variables[var + "_brightness_temperature_lut"][...]
                ncvar.set_auto_maskandscale(False)
                unscaled_rad = ncvar[...]
                nparr = btlut[unscaled_rad]
                add_to_xarray(
                    btvarname,
                    nparr,
                    xarrays[data_type],
                    dataset_masks,
                    data_type,
                    nparr.mask,
                )
                xarrays[data_type][btvarname].attrs["units"] = "Kelvin"

            if (
                var in BT_CHANNELS + REF_CHANNELS
                and required_chan(chans, [radvarname])
                and "radiance_add_offset" in ncvar.ncattrs()
            ):
                LOG.info(
                    "        Reading %s channel %s into RADIANCE variable %s",
                    data_type,
                    var,
                    radvarname,
                )

                ncvar.set_auto_maskandscale(True)

                # Bowtie correction
                nparr_bowtie = numpy.ma.masked_equal(ncvar[...].data, 65533)

                nparr_masked = numpy.ma.masked_greater(
                    ((ncvar[...] - ncvar.add_offset) / ncvar.scale_factor)
                    * ncvar.radiance_scale_factor
                    + ncvar.radiance_add_offset,
                    ncvar.valid_max * ncvar.radiance_scale_factor
                    + ncvar.radiance_add_offset,
                )

                add_to_xarray(
                    radvarname,
                    nparr_masked,
                    xarrays[data_type],
                    dataset_masks,
                    data_type,
                    nparr_bowtie.mask,
                )

                for attrname in ncvar.ncattrs():
                    xarrays[data_type][radvarname].attrs[attrname] = ncvar.getncattr(
                        attrname
                    )

            if var in REF_CHANNELS and required_chan(chans, [refvarname]):
                LOG.info(
                    "        Reading %s channel %s into REFLECTANCE variable %s",
                    data_type,
                    var,
                    refvarname,
                )

                nparr_bowtie = numpy.ma.masked_equal(ncvar[...].data, 65533)

                nparr_masked = numpy.ma.masked_greater(
                    ncvar[...], ncvar.valid_max * ncvar.scale_factor + ncvar.add_offset
                )

                # Gamma, Scale factor from old Vis product
                from geoips.data_manipulations.corrections import (
                    apply_gamma,
                    apply_scale_factor,
                )

                nparr_masked = apply_gamma(nparr_masked, 1.5)
                nparr_masked = apply_scale_factor(nparr_masked, 100)

                add_to_xarray(
                    refvarname,
                    nparr_masked,
                    xarrays[data_type],
                    dataset_masks,
                    data_type,
                    nparr_bowtie.mask,
                )

            for geo_target_data_type in geo_target_data_types:
                if var in GVARLIST[geo_target_data_type] and required_geo(
                    chans, geo_target_data_type
                ):
                    source_file_vars = tracked_data_type_vars[geo_target_data_type][
                        Start_date
                    ]
                    if geo_target_data_type == vis_data_type and not all(
                        [x in source_file_vars for x in vis_vars]
                    ):
                        # Only add geolocation to vis_data_type if corresponding
                        # channels are in the source file
                        continue
                    xvarname = var
                    if var in xvarnames:
                        xvarname = xvarnames[var]

                    if not required_geo_chan(xarrays, xvarname):
                        continue

                    LOG.info(
                        "        Reading %s geolocation channel %s into GEOLOCATION "
                        "variable %s",
                        geo_target_data_type,
                        var,
                        xvarname,
                    )
                    nparr = numpy.ma.masked_equal(ncvar[...], ncvar._FillValue)
                    add_to_xarray(
                        xvarname,
                        nparr,
                        xarrays[geo_target_data_type],
                        dataset_masks,
                        geo_target_data_type,
                        nparr.mask,
                    )
                    for attrname in ncvar.ncattrs():
                        xarrays[geo_target_data_type][xvarname].attrs[attrname] = (
                            ncvar.getncattr(attrname)
                        )

        # LongName is something like:
        # 'VIIRS/JPSS1 Day/Night Band 6-Min L1B Swath SDR 750m NRT'
        # or
        # 'VIIRS/NPP Day/Night Band 6-Min L1B Swath 750m'
        long_name_parts = ncdf_file.LongName.split(" ")
        if long_name_parts[-1] == "NRT":
            resolution_m = long_name_parts[-2]
        else:
            resolution_m = long_name_parts[-1]
        resolution_m = int(resolution_m.replace("m", ""))
        LOG.info("Resolution: %s", resolution_m)
        xarrays[data_type].attrs["resolution_km"] = resolution_m / 1000.0

        # close the files
        ncdf_file.close()

    # Geolocation resampling
    if resample and area_def:
        adname = area_def.area_id
        new_shape = area_def.shape

        LOG.info("")
        LOG.info("Getting geolocation information for {}.".format(adname))

        for dtype in xarrays.keys():
            if "latitude" not in xarrays[dtype].variables:
                LOG.info(
                    "No data read for dataset %s, removing from xarray list", dtype
                )
                continue
            fldk_lats = xarrays[dtype]["latitude"]
            fldk_lons = xarrays[dtype]["longitude"]
            # Get just the metadata we need
            geo_metadata = _get_geolocation_metadata(
                fldk_lats.shape, fnames, xarrays[dtype]
            )
            lines, samples = get_indexes(geo_metadata, fldk_lats, fldk_lons, area_def)

            index_mask = lines != -999

            new_dim0 = "dim_{:d}".format(new_shape[0])
            new_dim1 = "dim_{:d}".format(new_shape[1])

            # Resample the mask
            old_dataset_masks = dataset_masks[dtype]
            dataset_masks[dtype] = numpy.full(new_shape, True)
            dataset_masks[dtype][index_mask] = old_dataset_masks[
                lines[index_mask], samples[index_mask]
            ]

            for varname in xarrays[dtype].variables.keys():
                new_var = numpy.full(new_shape, -999.1)
                new_var[index_mask] = xarrays[dtype][varname].values[
                    lines[index_mask], samples[index_mask]
                ]
                if varname not in list(xvarnames.values()) + ["latitude", "longitude"]:
                    # Set values <= -999.9 to NaN so they also get interpolated.
                    new_var[numpy.where(new_var <= -999.9)] = numpy.nan
                    # Interpolate missing data.
                    out_var = numpy.array(pd.DataFrame(new_var).interpolate())
                    # Reset the mask so these values get shown.
                    skip_mask = numpy.where(
                        numpy.isnan(new_var) & ~numpy.isnan(out_var)
                    )
                    dataset_masks[dtype][skip_mask] = False
                else:
                    out_var = new_var
                xarrays[dtype][varname] = xr.DataArray(
                    out_var, dims=[new_dim0, new_dim1]
                )

        LOG.interactive("Done with geolocation for {}".format(adname))
        LOG.info("")

    # Clean up the final xarray dictionary
    xarray_returns = {}
    for dtype in xarrays.keys():
        if "latitude" not in xarrays[dtype].variables:
            LOG.info("No data read for dataset %s, removing from xarray list", dtype)
            continue
        for varname in xarrays[dtype].variables.keys():
            xarrays[dtype][varname] = xarrays[dtype][varname].where(
                ~dataset_masks[dtype]
            )
        if "DNBRad" in list(xarrays[dtype].variables.keys()) and required_chan(
            chans, ["DNBRef"]
        ):
            with import_optional_dependencies(loglevel="info"):
                """Attempt to import a package log to INFO if the import fails."""
                from lunarref.lib.liblunarref import lunarref

                lunarref_data = lunarref(
                    xarrays["DNB"]["DNBRad"].to_masked_array(),
                    xarrays["DNB"]["solar_zenith_angle"].to_masked_array(),
                    xarrays["DNB"]["lunar_zenith_angle"].to_masked_array(),
                    xarrays["DNB"].start_datetime.strftime("%Y%m%d%H"),
                    xarrays["DNB"].start_datetime.strftime("%M"),
                    xarrays["DNB"]["moon_phase_angle"].mean(),
                )
                lunarref_data = numpy.ma.masked_less_equal(
                    lunarref_data, -999, copy=False
                )
                xarrays["DNB"]["DNBRef"] = xr.DataArray(
                    lunarref_data, dims=xarrays["DNB"]["DNBRad"].dims
                )
        # This will not duplicate memory - reference
        xarray_returns[dtype] = xarrays[dtype]

    # Force masking of latitude/longitude fill values, otherwise causes issues
    # when sectoring
    # fill_value = -999.9
    for dtype in xarray_returns:
        bad_llmask = xarray_returns[dtype]["latitude"] == -999.9
        for var in xarray_returns[dtype].variables.keys():
            xarray_returns[dtype][var] = xarray_returns[dtype][var].where(~bad_llmask)

    xarray_returns["METADATA"] = list(xarray_returns.values())[0][[]]

    return xarray_returns



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""matplotlib utilities."""

# Python Standard Libraries
import logging
import matplotlib
import matplotlib.pyplot as plt

from geoips.filenames.base_paths import PATHS as gpaths
from geoips.interfaces import title_formatters
from geoips.image_utils.maps import draw_features, draw_gridlines

matplotlib.use("agg")
rc_params = matplotlib.rcParams
LOG = logging.getLogger(__name__)


def percent_unmasked_rgba(rgba):
    """
    Convert to percent.

    Return percentage of array that is NOT fully transparent / masked
    (ie, non-zero values in the 4th dimension)

    Parameters
    ----------
    rgba : numpy.ndarray
        4 dimensional array where the 4th dimension is the alpha transparency array:
        1 is fully opaque, 0 is fully transparent

    Returns
    -------
    coverage : float
        Coverage in percentage, between 0 and 100.
    """
    import numpy

    coverage = 100.0 * numpy.count_nonzero(rgba[:, :, 3]) / rgba[:, :, 3].size
    return coverage


def rgba_from_arrays(red, grn, blu, alp=None):
    """
    Return rgba from red, green, blue, and alpha arrays.

    Parameters
    ----------
    red : numpy.ndarray
        red gun values
    grn : numpy.ndarray
        green gun values
    blu : numpy.ndarray
        blue gun values
    alp : numpy.ndarray, optional
        alpha values 1 is fully opaque, 0 is fully transparent
        If none, calculate alpha from red, grn, blu guns

    Returns
    -------
    rgba : numpy.ndarray
        4 layer dimensional numpy.ndarray
    """
    import numpy

    if alp is None:
        alp = alpha_from_masked_arrays([red, grn, blu])
    red.fill_value = 0
    grn.fill_value = 0
    blu.fill_value = 0
    rgba = numpy.dstack([red.filled(), grn.filled(), blu.filled(), alp])
    return rgba


def alpha_from_masked_arrays(arrays):
    """
    Convert from arrays to alpha.

    Return an alpha transparency array based on the masks from a list of
    masked arrays. 0=transparent, 1=opaque

    Parameters
    ----------
    arrays : numpy.ndarray
        list of numpy masked arrays, must all be the same shape

    Returns
    -------
    alp : numpy.ndarray
        the alpha transparency layer in matplotlib, values between
        0 and 1, where 0 is fully transparent and 1 is fully opaque
    """
    import numpy

    alp = numpy.zeros(arrays[0].shape, dtype=bool)
    for img in arrays:
        try:
            if img.mask is not numpy.False_:
                alp += img.mask
        except AttributeError:
            pass
    # You will get yelled at by numpy if you removed the "alp.dtype" portion of this.
    #   It thinks you are trying to cast alp to be an integer.
    alp = numpy.array(alp, dtype=float)
    alp -= float(1)
    alp *= float(-1)
    return alp


def plot_overlays(
    mapobj,
    curr_ax,
    area_def,
    feature_annotator=None,
    gridline_annotator=None,
    features_zorder=None,
    gridlines_zorder=None,
):
    """
    Plot specified coastlines and gridlines on the matplotlib axes.

    Parameters
    ----------
    mapobj : map object
        Basemap or CRS object for boundary and gridline plotting.
    ax : matplotlib.axes._axes.Axes
        matplotlib Axes object for boundary and gridline plotting.
    area_def : AreaDefinition
        pyresample AreaDefinition object specifying the area covered by
        the current plot
    feature_annotator : YamlPlugin
        A feature annotator plugin instance.
    gridline_annotator : YamlPlugin
       A gridlines annotator plugin instance.
    """
    draw_features(mapobj, curr_ax, feature_annotator, zorder=features_zorder)
    draw_gridlines(
        mapobj, area_def, curr_ax, gridline_annotator, zorder=gridlines_zorder
    )


def save_image(
    fig,
    out_fname,
    is_final=True,
    image_datetime=None,
    remove_duplicate_minrange=None,
    savefig_kwargs=None,
):
    """
    Save the image specified by the matplotlib figure "fig" to the filename out_fname.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        Figure object that needs to be written to a file.
    out_fname : str
        full path to the output filename
    is_final : bool, default=True
        Final imagery must set_axis_on for all axes. Non-final imagery
        must be transparent with set_axis_off for all axes, and no pad
        inches.

    Notes
    -----
    No return values (image is written to disk and IMAGESUCCESS is
    written to log file)
    """
    rc_params = matplotlib.rcParams
    from os.path import dirname, exists as pathexists
    from geoips.filenames.base_paths import make_dirs

    if savefig_kwargs is None:
        savefig_kwargs = {}

    if is_final:
        if not pathexists(dirname(out_fname)):
            make_dirs(dirname(out_fname))
        for ax in fig.axes:
            LOG.info("Adding ax to %s", ax)
            ax.set_axis_on()
        # final with titles, labels, etc.
        # Note bbox_inches='tight' removes white space, pad_inches=0.1 puts back in
        # a bit of white space.
        LOG.interactive("Writing %s", out_fname)
        fig.savefig(
            out_fname,
            dpi=rc_params["figure.dpi"],
            pad_inches=0.1,
            bbox_inches="tight",
            transparent=False,
        )
        if remove_duplicate_minrange is not None:
            remove_duplicates(out_fname, remove_duplicate_minrange)
    else:
        # Get rid of the colorbar axis for non-final imagery
        if not pathexists(dirname(out_fname)):
            make_dirs(dirname(out_fname))
        # no annotations
        # frameon=False makes it have no titles / lat/lons. does not avoid
        # colorbar, since that is its own ax
        for ax in fig.axes:
            LOG.info("Removing ax from %s", ax)
            ax.set_axis_off()

        LOG.info("Writing %s", out_fname)

        fig.savefig(
            out_fname,
            dpi=rc_params["figure.dpi"],
            pad_inches=0.0,
            transparent=True,
            facecolor="none",
            **savefig_kwargs,
        )
        if remove_duplicate_minrange is not None:
            remove_duplicates(out_fname, remove_duplicate_minrange)

    LOG.info("IMAGESUCCESS wrote %s", out_fname)
    if image_datetime is not None:
        from datetime import datetime

        LOG.info("LATENCY %s %s", datetime.utcnow() - image_datetime, out_fname)
    return [out_fname]


def remove_duplicates(fname, min_range):
    """Not implemented."""
    pass


def get_title_string_from_objects(
    area_def,
    xarray_obj,
    product_name_title,
    product_datatype_title=None,
    bg_xarray=None,
    bg_product_name_title=None,
    bg_datatype_title=None,
    title_copyright=None,
    title_formatter=None,
):
    """
    Get the title from object information.

    Parameters
    ----------
    area_def : AreaDefinition
        pyresample AreaDefinition object specifying the area covered by
        the current plot
    xarray_obj : xarray.Dataset
        data used to produce product
    product_name_title : str
        name to display for the title
    product_datatype_title : str, optional
        the data type
    bg_xarray : xarray, optional
        data used for background
    bg_product_name_title : str, optional
        background product title
    bg_datatype_title : str, optional
        background data type
    title_copyright : str, optional
        string for copyright
    title_formatter : str, optional
        format for title

    Returns
    -------
    title_string : str
        the title to use for matplotlib
    """
    if title_copyright is None:
        title_copyright = gpaths["GEOIPS_COPYRIGHT"]

    from geoips.sector_utils.utils import is_sector_type

    if product_datatype_title is None:
        product_source_name = xarray_obj.source_name
        product_platform_name = xarray_obj.platform_name
        if "model" in product_name_title:
            try:
                product_source_name = xarray_obj.model_source
                product_platform_name = ""
            except AttributeError:
                LOG.info("No model source in xarray")
        product_datatype_title = "{0} {1}".format(
            product_platform_name.upper(), product_source_name.upper()
        )

    if bg_xarray is not None and bg_datatype_title is None:
        bg_datatype_title = "{0} {1}".format(
            bg_xarray.platform_name.upper(), bg_xarray.source_name.upper()
        )

    if title_formatter is not None:
        title_formatter_plugin = title_formatters.get_plugin(title_formatter)
    elif is_sector_type(area_def, "tc"):
        title_formatter_plugin = title_formatters.get_plugin("tc_standard")
    else:
        title_formatter_plugin = title_formatters.get_plugin("static_standard")

    title_string = title_formatter_plugin(
        area_def,
        xarray_obj,
        product_name_title,
        product_datatype_title=product_datatype_title,
        bg_xarray=bg_xarray,
        bg_product_name_title=bg_product_name_title,
        bg_datatype_title=bg_datatype_title,
        title_copyright=title_copyright,
    )

    return title_string


def plot_image(main_ax, data, mapobj, mpl_colors_info, zorder=None, bkgrnd_clr=None):
    """
    Plot the "data" array and map in the matplotlib "main_ax".

    Parameters
    ----------
    main_ax : Axes
        matplotlib Axes object for plotting data and overlays
    data : numpy.ndarray)
        Numpy array of data to plot
    mapobj : Map Object
        Basemap or Cartopy CRS instance
    mpl_colors_info : dict
        Specifies matplotlib Colors parameters for use in both plotting and colorbar
    bkgrnd_clr: string (default = None)
        - The background color to apply to the image. If None, the image background will
          be white.

    See Also
    --------
    geoips.image_utils.mpl_utils.create_colorbar
        for field descriptions for mpl_colors_info
    """
    # main_ax.set_aspect('auto')
    if bkgrnd_clr:
        main_ax.set_facecolor(bkgrnd_clr)

    LOG.info("imshow")
    import numpy
    from geoips.image_utils.maps import is_crs

    if is_crs(mapobj):
        # Apparently cartopy handles the flipud

        # NOTE passing zorder=None is NOT the same as not passing zorder.
        # Ensure if zorder is passed, it is propagated to imshow.
        extra_args = {}
        if zorder is not None:
            extra_args = {"zorder": zorder}
        main_ax.imshow(
            data,
            transform=mapobj,
            extent=mapobj.bounds,
            cmap=mpl_colors_info["cmap"],
            norm=mpl_colors_info["norm"],
            **extra_args,
        )
    else:
        mapobj.imshow(
            numpy.flipud(data),
            ax=main_ax,
            cmap=mpl_colors_info["cmap"],
            norm=mpl_colors_info["norm"],
            aspect="auto",
        )

    return mapobj


def create_figure_and_main_ax_and_mapobj(
    x_size,
    y_size,
    area_def,
    frame_clr=None,
    font_size=None,
    existing_mapobj=None,
    noborder=False,
):
    """
    Create a figure of x pixels horizontally and y pixels vertically.

    Use information from matplotlib.rcParams.

    Parameters
    ----------
    x_size : int
        number pixels horizontally
        xsize = (float(x_size)/dpi)/(right_margin - left_margin)
    y_size : int
        number pixels vertically
        ysize = (float(y_size)/dpi)/(top_margin - bottom_margin)
    area_def : AreaDefinition
        pyresample AreaDefinition object - used for
        initializing map object (basemap or cartopy)
    frame_clr: string (default = None)
        - The color to apply to the frame of the image (where title, label, colorbar...)
          is set. If None, the frame background color will be white.
    font_size : int
        matplotlib font size
    existing_mapobj : CRS or basemap, optional
        If specified, do not regenerate mapobj. If None, create
        CRS or basemap object from specified area_def.
    noborder : bool, default=False
        If true, use [0, 0, 1, 1] for axes (allowing for image exact
        shape of sector).

    Returns
    -------
    fig : matplotlib.figure.Figure
        matplotlib Figure object to subsequently use for plotting
        imagery / colorbars / etc
    main_ax : matplotlib.axes._axes.Axes
        matplotlib Axes object corresponding to the single main plotting area.
    mapobj : mapobject
        cartopy crs or Basemap object for plotting
    """
    import matplotlib

    matplotlib.use("agg")
    rc_params = matplotlib.rcParams

    set_fonts(y_size, font_size=font_size)

    # Gather needed rcParam constants
    dpi = rc_params["figure.dpi"]

    # I can't seem to get a clean image with no border unless the ax fills the
    # whole figure.
    # Titles / labels don't show up unless we use figure.subplot rc_params
    # It does not appear to be possible to plot the image only once for both
    # the clean and annotated image.
    if noborder:
        left_margin = 0.0
        right_margin = 1.0
        bottom_margin = 0.0
        top_margin = 1.0
    else:
        left_margin = rc_params[
            "figure.subplot.left"
        ]  # Fractional distance from left edge of figure for subplot
        right_margin = rc_params[
            "figure.subplot.right"
        ]  # Fractional distance from right edge of figure for subplot
        bottom_margin = rc_params[
            "figure.subplot.bottom"
        ]  # Fractional distance from bottom edge of figure for subplot
        top_margin = rc_params[
            "figure.subplot.top"
        ]  # Fractional distance from top edge of figure for subplot

    xsize = (float(x_size) / dpi) / (right_margin - left_margin)
    ysize = (float(y_size) / dpi) / (top_margin - bottom_margin)

    if existing_mapobj is None:
        LOG.info("creating mapobj instance")
        mapobj = area_def.to_cartopy_crs()
    else:
        LOG.info("mapobj already exists, not recreating")
        mapobj = existing_mapobj

    LOG.info(
        "Creating figure: left, right, bottom, top, xsize, ysize %s %s %s %s %s %s",
        left_margin,
        right_margin,
        bottom_margin,
        top_margin,
        xsize,
        ysize,
    )

    from geoips.image_utils.maps import is_crs

    if is_crs(mapobj):
        # frameon=False creates transparent titles with cartopy
        fig = plt.figure(facecolor=frame_clr)
    else:
        fig = plt.figure(frameon=False, facecolor=frame_clr)
    fig.set_size_inches(xsize, ysize)
    set_fonts(y_size, font_size=font_size)

    LOG.info(
        "Creating main ax: left, bottom, width, height %s %s %s %s",
        left_margin,
        bottom_margin,
        right_margin - left_margin,
        top_margin - bottom_margin,
    )

    # cartopy figure with non-transparent titles
    # In [71]: fig = plt.figure()
    # In [72]: main_ax = fig.add_axes([0, 0, 1, 1], projection=mapobj, frame_on=True)
    # In [73]: plt.imshow(plot_data, transform=mapobj, extent=mapobj.bounds)
    # Out[73]: <matplotlib.image.AxesImage at 0x7fcfe5f75eb0>
    # In [74]: main_ax.set_title('Hello', y=1.2, fontsize=20)
    # Out[74]: Text(0.5, 1.2, 'Hello')
    # In [75]: fig.savefig('test.png', transparent=False)

    if is_crs(mapobj):
        # Transparent behind titles without frame_on True
        main_ax = fig.add_axes(
            [
                left_margin,
                bottom_margin,
                right_margin - left_margin,
                top_margin - bottom_margin,
            ],
            projection=mapobj,
            frame_on=not noborder,
        )
    else:
        main_ax = plt.Axes(
            fig,
            [
                left_margin,
                bottom_margin,
                right_margin - left_margin,
                top_margin - bottom_margin,
            ],
        )
    main_ax.set_axis_off()
    fig.add_axes(main_ax)

    return fig, main_ax, mapobj


def set_fonts(figure_y_size, font_size=None):
    """
    Set the fonts in the matplotlib.rcParams dictionary, using matplotlib.rc.

    Parameters
    ----------
    figure_y_size : int
        Font size set relative to number of pixels in the y direction
    """
    import matplotlib

    matplotlib.use("agg")
    mplrc = matplotlib.rc

    # Update font size based on number of lines
    if font_size is not None:
        title_fsize = font_size
    elif int(figure_y_size) / 1000 != 0:
        title_fsize = 20 * int(figure_y_size) / 1000
    else:
        title_fsize = 20

    font = {"family": "sans-serif", "weight": "bold", "size": title_fsize}

    LOG.info("Setting font size to %s for figure_y_size %s", title_fsize, figure_y_size)
    mplrc("font", **font)


def set_title(ax, title_string, figure_y_size, xpos=None, ypos=None, fontsize=None):
    """
    Set the title on figure axis "ax" to string "title_string".

    Parameters
    ----------
    ax : Axes
        matplotlib.axes._axes.Axes object to add the title
    title_string : str
        string specifying title to attach to axis "ax"
    figure_y_size : int
        vertical size of the image, used to proportionally set the title size
    xpos : float, optional
        x position of the title
    ypos : float, optional
        y position of the title
    fontsize : int, optional
        matplotlib font size
    """
    import matplotlib

    matplotlib.use("agg")
    rc_params = matplotlib.rcParams

    # Provide pad room between characters
    fontspace = rc_params["font.size"]
    title_line_space = float(fontspace) / figure_y_size
    # num_title_lines = len(title_string.split('\n'))

    if xpos is None:
        xpos = 0.5  # This centers the title
    if ypos is None:
        # ypos = 1 + title_line_space*2
        ypos = (
            1 + title_line_space * 2
        )  # This is relative to main_ax, so greater than 1.
    LOG.info(
        "Setting title: font size %s, xpos %s ypos %s, title_line_space %s",
        fontspace,
        xpos,
        ypos,
        title_line_space,
    )
    LOG.info("    Title string: %s", title_string)

    # from cartopy.mpl.geoaxes import GeoAxes # isinstance(ax, GeoAxes)
    # Slightly more space between title and lon labels
    # ax.set_title(title_string, y=ypos, fontsize=fontsize)
    ax.set_title(title_string, position=[xpos, ypos], fontsize=fontsize)


def create_colorbar(fig, mpl_colors_info):
    """
    Routine to create a single colorbar.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        Figure object to attach the colorbar - the colorbar will create its own ax
    mpl_colors_info : dict
        Dictionary of matplotlib Color information, required fields in Notes below.

    Returns
    -------
    cbar : matplotlib.colorbar.Colorbar
        This will have all the pertinent information for ensuring plot and
        colorbar use the same parameters

    Notes
    -----
    Required mpl_colors_info fields::

        mpl_colors_info['cmap'] (Colormap):
            matplotlib.colors.Colormap object (LinearSegmentedColormap, etc)
            this is used to plot the image and to generated the colorbar
        mpl_colors_info['norm'] (Normalize):
            matplotlib.colors.Normalize object (BoundaryNorm, Normalize, etc)
            again, this should be used to plot the data also
        mpl_colors_info['cbar_ticks'] (list):
            list of floats - values requiring tick marks on the colorbar
        mpl_colors_info['cbar_tick_labels'] (list)
            list of values to use to label tick marks, if other than
            found in cbar_ticks
        mpl_colors_info['boundaries'] (list):
            if cmap_norm is BoundaryNorm, list of boundaries for discrete colors
        mpl_colors_info['cbar_spacing'] (string):
            DEFAULT 'proportional', 'uniform' or 'proportional'
        mpl_colors_info['cbar_label'] (string):
            string label for colorbar
        mpl_colors_info['colorbar']: (bool)
            True if a colorbar should be included in the image, False if no cbar


    Colorbar set as::

        cbar_ax = fig.add_axes([<cbar_start_pos>, <cbar_bottom_pos>,
                                <cbar_width>, <cbar_height>])
        cbar = fig.colorbar(mappable=matplotlib.cm.ScalarMappable(norm=cmap_norm,
                                                                  cmap=mpl_cmap),
                            cax=cbar_ax,
                            norm=cmap_norm,
                            boundaries=cmap_boundaries,
                            spacing=cbar_spacing,
                            **cbar_kwargs)
        cbar.set_ticks(cbar_ticks, labels=cbar_tick_labels, **set_ticks_kwargs)
        if cbar_label:
            cbar.set_label(cbar_label, **set_label_kwargs)
    """
    # Required - needed for both colormaps and colorbars.
    cmap_norm = mpl_colors_info["norm"]
    mpl_cmap = mpl_colors_info["cmap"]
    cmap_boundaries = mpl_colors_info["boundaries"]

    # Required, for setting colorbar labels with set_label and ticks with set_ticks
    cbar_label = mpl_colors_info["cbar_label"]
    cbar_ticks = mpl_colors_info["cbar_ticks"]
    if cbar_ticks is None:
        cbar_ticks = [cmap_norm.vmin, cmap_norm.vmax]

    # Optional - can be specified in colorbar_kwargs
    cbar_spacing = "proportional"
    if mpl_colors_info.get("cbar_spacing") is not None:
        cbar_spacing = mpl_colors_info["cbar_spacing"]

    # Optional - can be specified in set_ticks_kwargs. None will be mapped to
    # cbar_ticks.
    cbar_tick_labels = None
    if "cbar_tick_labels" in mpl_colors_info:
        cbar_tick_labels = mpl_colors_info["cbar_tick_labels"]
    if mpl_colors_info["cbar_tick_labels"] is None:
        cbar_tick_labels = cbar_ticks

    # Allow arbitrary kwargs to colorbar, but ensure our defaults for extend
    # and orientation are set.
    cbar_kwargs = {}
    if mpl_colors_info.get("colorbar_kwargs") is not None:
        cbar_kwargs = mpl_colors_info["colorbar_kwargs"].copy()
    if "extend" not in cbar_kwargs:
        cbar_kwargs["extend"] = "both"
    if "spacing" not in cbar_kwargs:
        cbar_kwargs["spacing"] = cbar_spacing
    if "orientation" not in cbar_kwargs:
        cbar_kwargs["orientation"] = "horizontal"

    # Allow arbitrary kwargs to set_ticks, but ensure our defaults for labels
    # and font sizes are set
    set_ticks_kwargs = {}
    if mpl_colors_info.get("set_ticks_kwargs") is not None:
        set_ticks_kwargs = mpl_colors_info["set_ticks_kwargs"].copy()
    if "size" not in set_ticks_kwargs:
        set_ticks_kwargs["size"] = "small"
    if "labels" not in set_ticks_kwargs or set_ticks_kwargs["labels"] is None:
        set_ticks_kwargs["labels"] = cbar_tick_labels

    # Allow arbitrary kwargs to set_label
    set_label_kwargs = {}
    if mpl_colors_info.get("set_label_kwargs") is not None:
        set_label_kwargs = mpl_colors_info["set_label_kwargs"].copy()
    if "size" not in set_label_kwargs:
        set_label_kwargs["size"] = rc_params["font.size"]

    left_margin = rc_params[
        "figure.subplot.left"
    ]  # Fractional distance from left edge of figure for subplot
    right_margin = rc_params[
        "figure.subplot.right"
    ]  # Fractional distance from left edge of figure for subplot

    # Optional positioning information for the colorbar axis - default
    # location appropriately
    if mpl_colors_info.get("cbar_ax_left_start_pos") is not None:
        ax_left_start_pos = mpl_colors_info["cbar_ax_left_start_pos"]
    else:
        ax_left_start_pos = 2 * left_margin
        if mpl_colors_info.get("cbar_full_width") is True:
            ax_left_start_pos = left_margin  # Full width colorbar

    ax_bottom_start_pos = 0.05
    if mpl_colors_info.get("cbar_ax_bottom_start_pos") is not None:
        ax_bottom_start_pos = mpl_colors_info["cbar_ax_bottom_start_pos"]

    if mpl_colors_info.get("cbar_ax_width") is not None:
        ax_width = mpl_colors_info["cbar_ax_width"]
    else:
        ax_width = 1 - 4 * left_margin
        if mpl_colors_info.get("cbar_full_width") is True:
            ax_width = right_margin - left_margin  # Full width colorbar

    ax_height = 0.020
    if mpl_colors_info.get("cbar_ax_height") is not None:
        ax_height = mpl_colors_info["cbar_ax_height"]

    # Note - if we want to support "automated" pyplot.colorbar placement,
    # may need a method to "turn off" cbar_ax
    # [left, bottom, width, height]
    cbar_ax = fig.add_axes(
        [ax_left_start_pos, ax_bottom_start_pos, ax_width, ax_height]
    )
    cbar = plt.colorbar(
        mappable=matplotlib.cm.ScalarMappable(norm=cmap_norm, cmap=mpl_cmap),
        cax=cbar_ax,
        norm=cmap_norm,
        boundaries=cmap_boundaries,
        **cbar_kwargs,
    )

    if cbar_ticks:
        # matplotlib 3.6.0 sometimes has inconsistent results with including
        # minor ticks or not.
        # Unclear why it impacts some colorbars and not others.
        # We may eventually add support for including minor ticks within
        # mpl_colors_info, but for now explicitly turn off minor ticks so
        # outputs will continue to match (use the old default).
        cbar.minorticks_off()
        cbar.set_ticks(cbar_ticks, **set_ticks_kwargs)
    if cbar_label:
        cbar.set_label(cbar_label, **set_label_kwargs)

    return cbar



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Data manipulation steps for "airmass" EUMETSAT RGB product.

This algorithm expects four Infrared channels for an RGB image:
* Red SEVIRI B05BT - B06BT
* Green SEVIRI B08BT - B09BT
* Blue SEVIRI B05BT
"""
import logging

LOG = logging.getLogger(__name__)

interface = "algorithms"
family = "xarray_to_numpy"
name = "airmass"


def call(xobj):
    """Airmass product algorithm data manipulation steps.

    This algorithm expects TBs from four SEVIRI channels:

    * Red: B05BT - B06BT
    * Green: B08BT - B09BT
    * Blue: B05BT

    Parameters
    ----------
    arrays : list of numpy.ndarray
        * list of numpy.ndarray or numpy.MaskedArray of channel data,
            in order of sensor "channels" list
        * Degrees Kelvin

    Returns
    -------
    numpy.ndarray
        numpy.ndarray or numpy.MaskedArray of qualitative RGBA image output
    """
    bt05 = xobj["B05BT"].to_masked_array()  # 6.2 um TBs, B05 SEVIRI
    bt06 = xobj["B06BT"].to_masked_array()  # 7.3 um TBs, B06 SEVIRI
    bt08 = xobj["B08BT"].to_masked_array()  # 9.7 um TBs, B08 SEVIRI
    bt09 = xobj["B09BT"].to_masked_array()  # 10.8 um TBs, B09 SEVIRI

    # Convert TB from Kevin to Celsius
    from geoips.data_manipulations.conversions import unit_conversion

    bt05 = unit_conversion(bt05, input_units="Kelvin", output_units="celsius")
    bt06 = unit_conversion(bt06, input_units="Kelvin", output_units="celsius")
    bt08 = unit_conversion(bt08, input_units="Kelvin", output_units="celsius")
    bt09 = unit_conversion(bt09, input_units="Kelvin", output_units="celsius")

    from geoips.data_manipulations.corrections import apply_data_range, apply_gamma

    data_range = [-25, 0]
    red = apply_data_range(
        bt05 - bt06,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )  # need inverse option?
    red = apply_gamma(red, 1.0)
    red = apply_gamma(red, 1.0)

    data_range = [-40, 5]
    grn = apply_data_range(
        bt08 - bt09,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )
    grn = apply_gamma(grn, 1.0)
    grn = apply_gamma(grn, 1.0)

    data_range = [-65, -30]
    blu = apply_data_range(
        bt05,
        min_val=data_range[0],
        max_val=data_range[1],
        min_outbounds="crop",
        max_outbounds="crop",
        norm=True,
        inverse=False,
    )  # create image of deep clouds in blueish color
    # norm=True, inverse=False)    #create image of deep clouds in yellowish color
    blu = apply_gamma(blu, 1.0)
    blu = apply_gamma(blu, 1.0)

    from geoips.image_utils.mpl_utils import alpha_from_masked_arrays, rgba_from_arrays

    alp = alpha_from_masked_arrays([red, grn, blu])
    rgba = rgba_from_arrays(red, grn, blu, alp)

    return rgba



"""Unit test module for the CLI command 'geoips config delete-registries'."""

import pytest

from geoips.commandline.commandline_interface import main
from geoips.errors import PluginRegistryError
from geoips.plugin_registry import plugin_registry
from tests.unit_tests.commandline.cli_top_level_tester import (
    generate_id,
    check_valid_command_using_monkeypatch,
)

valid_args = [
    ["geoips", "config", "delete-registries"],
    ["geoips", "config", "delete-registries", "-p", "geoips"],
    ["geoips", "config", "delete-registries", "-n", "geoips.plugin_packages"],
    [
        "geoips",
        "config",
        "delete-registries",
        "-n",
        "geoips.plugin_packages",
        "-p",
        "geoips",
    ],
    # Used to rebuild the registry at the end of testing.
    ["creating-registries"],
]
valid_expected = [
    {
        "log_level": "interactive",
        "command": "geoips config delete-registries",
        "namespace": "geoips.plugin_packages",
        "packages": None,
    },
    {
        "log_level": "interactive",
        "command": "geoips config delete-registries",
        "namespace": "geoips.plugin_packages",
        "packages": ["geoips"],
    },
    {
        "log_level": "interactive",
        "command": "geoips config delete-registries",
        "namespace": "geoips.plugin_packages",
        "packages": None,
    },
    {
        "log_level": "interactive",
        "command": "geoips config delete-registries",
        "namespace": "geoips.plugin_packages",
        "packages": ["geoips"],
    },
    # Dummy dict needed for registry rebuild
    {"empty": "dict"},
]
invalid_args = [
    ["geoips", "config", "delete-registries", "-p", "fake_package"],
    ["geoips", "config", "delete-registries", "-n", "fake.plugin_packages"],
]


@pytest.mark.parametrize("cli_args", [args for args in invalid_args], ids=generate_id)
def test_delete_registries_invalid(monkeypatch, cli_args):
    """Test the CLI command 'geoips config delete-registries' using invalid args.

    Parameters
    ----------
    cli_args: list[str]
        - A list of strings representing the CLI command being ran.
    """
    monkeypatch.setattr("sys.argv", cli_args)
    with pytest.raises(PluginRegistryError):
        main()


@pytest.mark.parametrize(
    "cli_args,expected",
    [(args, expected) for args, expected in zip(valid_args, valid_expected)],
    ids=generate_id,
)
def test_delete_registries_valid(monkeypatch, cli_args, expected):
    """Test the CLI command 'geoips config delete-registries' using valid args.

    Parameters
    ----------
    cli_args: list[str]
        - A list of strings representing the CLI command being ran.
    expected: dict
        - A dictionary of expected argument values that are returned from the command
          being ran from args.
    """
    # Reset the registry at the end of testing. This could break other unit tests if the
    # registries are not rebuilt.
    if cli_args == ["creating-registries"]:
        plugin_registry.create_registries()
        return
    check_valid_command_using_monkeypatch(monkeypatch, cli_args, expected)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utility for reading HRIT datasets."""

import os
import re
import codecs
import logging
import shutil
import operator
from functools import reduce
from copy import copy
from struct import unpack
from datetime import datetime, timedelta

from geoips.utils.context_managers import import_optional_dependencies

with import_optional_dependencies(loglevel="interactive"):
    """Attempt to import xRITDecompress from pyPublicDecompWT.

    Needed to decompress xRIT data types.
    """
    from pyPublicDecompWT import xRITDecompress

import numpy as np

log = logging.getLogger(__name__)

# interface = None indicates to the GeoIPS interfaces that this is not a valid
# plugin, and this module will not be added to the GeoIPS plugin registry.
# This allows including python modules within the geoips/plugins directory
# that provide helper or utility functions to the geoips plugins, but are
# not full GeoIPS plugins on their own.
interface = None


class HritDtype(object):
    """HRIT data type."""

    types = {
        "uint8": "u1",
        "uint16": ">u2",
        "uint32": ">u4",
        "uint64": ">u8",
        "int8": "i1",
        "int16": ">i2",
        "int32": ">i4",
        "int64": ">i8",
        "byte": "B",
        "unicode": "U",
    }

    def __getattr__(self, type_name):
        """Get hritdtype attr."""
        try:
            return np.dtype(self.types[type_name])
        except KeyError:
            if type_name[0:4] == "char":
                if len(type_name) > 4:
                    return np.dtype(">S{}".format(type_name[4:]))
                else:
                    return np.dtype(">S")
            raise HritError("Unknown data type: {}".format(type_name))


dtype = HritDtype()


def read10bit(buff):
    """Read 10 bit little endian data from a buffer.

    Returns
    -------
    int
        16 bit unsigned int.
    """
    while True:
        b = buff.read(5)
        if not len(b):
            break
        n = int(codecs.encode(b, "hex"), 16)
        p3 = n & 0x3FF
        n >>= 10
        p2 = n & 0x3FF
        n >>= 10
        p1 = n & 0x3FF
        n >>= 10
        p0 = n & 0x3FF
        yield p0
        yield p1
        yield p2
        yield p3


class HritError(Exception):
    """Raise exception when errors occur in reading xRIT data files."""

    def __init__(self, msg, code=None):
        """Initialize HritError."""
        self.code = code
        self.value = msg

    def __str__(self):
        """Hriterror str method."""
        if self.code:
            return "{}: {}".format(self.code, self.value)
        else:
            return self.value


class HritFile(object):
    """Hrit File class."""

    _file_type_map = {
        0: "image",
        1: "gts",
        2: "text",
        3: "encryptionKey",
        128: "prologue",
        129: "epilogue",
    }

    def __init__(self, fname):
        """Initialize HRIT File object."""
        # Test for file
        if not os.path.isfile(fname):
            raise IOError("No such file or directory: {}".format(fname))

        self._name = os.path.abspath(fname)
        self._dirname = os.path.dirname(self.name)
        self._basename = os.path.basename(self.name)

        # 20201207 'rb' required for binary read in Python 3
        self._fobj = open(self._name, "rb")

        # Construct the metadata block map
        # Some of these are static, others change from file to file
        self._block_map = {
            0: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                ("type_code", dtype.uint8, 1),
                ("header_length", dtype.uint32, 1),
                ("data_length", dtype.uint64, 1),
            ],
            1: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                ("bits_per_pixel", dtype.uint8, 1),
                ("num_samples", dtype.uint16, 1),
                ("num_lines", dtype.uint16, 1),
                ("compression", dtype.uint8, 1),
            ],
            2: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                ("projection", dtype.char32, 1),
                ("sample_scale", dtype.int32, 1),
                ("line_scale", dtype.int32, 1),
                ("sample_offset", dtype.int32, 1),
                ("line_offset", dtype.int32, 1),
            ],
            3: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                (
                    "data_definition",
                    getattr(dtype, "char{}".format(self.__block_len_minus_two(3))),
                    1,
                ),
            ],
            4: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                (
                    "annotation",
                    getattr(dtype, "char{}".format(self.__block_len_minus_two(4))),
                    1,
                ),
            ],
            5: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                (
                    "time",
                    getattr(dtype, "char{}".format(self.__block_len_minus_two(5))),
                    1,
                ),
            ],  # Unsure how to convert
            6: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                (
                    "ancillary_text",
                    getattr(dtype, "char{}".format(self.__block_len_minus_two(6))),
                    1,
                ),
            ],
            7: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                (
                    "key_header_info",
                    getattr(dtype, "char{}".format(self.__block_len_minus_two(7))),
                    1,
                ),
            ],
            128: [
                ("block_num", dtype.uint8, 1),
                ("block_length", dtype.uint16, 1),
                ("sc_id", dtype.uint16, 1),
                ("chan_id", dtype.uint8, 1),
                ("segment", dtype.uint16, 1),
                ("start_segment", dtype.uint16, 1),
                ("end_segment", dtype.uint16, 1),
                ("data_field_repr", dtype.uint8, 1),
            ],
            129: None,
        }

        # Read the metadata itself
        self._metadata = self.__read_metadata()

    @property
    def block_info(self):
        """Block info."""
        if not hasattr(self, "_block_info"):
            self._block_info = self.__read_metadata_block_info()
        return self._block_info

    def __block_len_minus_two(self, block_num):
        """
        Return the length of a header block minus 2.

        This is for reading to the end of the block. This may seem silly,
        but is mainly used to make the code above readable.
        """
        try:
            return self.block_info[block_num][1] - 2
        except KeyError:
            return 0

    @property
    def block_map(self):
        """Return block map."""
        return self._block_map

    @property
    def metadata(self):
        """Return metadata."""
        if not hasattr(self, "_metadata"):
            self._metadata = self._read_metadata()
        return self._metadata

    # @property
    # def full_disk_geolocation_metadata(self):
    #     if not hasattr(self, '_full_disk_geolocation_metadata'):
    #         geomet = {}
    #         geomet['ob_area'] = re.sub(r'\W+', '',
    #                                            self.metadata['block_2']['projection'])
    #         geomet['num_lines'] = self.metadata['block_1']['num_lines']
    #         geomet['num_samples'] = self.metadata['block_1']['num_samples']
    #         geomet['line_scale'] = self.metadata['block_2']['line_scale']
    #         geomet['sample_scale'] = self.metadata['block_2']['sample_scale']
    #         geomet['line_offset'] = geomet['num_lines'] / 2
    #         geomet['sample_offset'] = geomet['num_samples'] / 2
    #         geomet['projection'] = self.metadata['block_2']['projection']
    #         geomet['start_datetime'] = self.start_datetime
    #         # Grabs the sub point longitude from the projection in the form
    #         #   proj(sublon)
    #         geomet['sublon'] = float(re.search(r"\((.+)\)",
    #                                  self.metadata['block_2']['projection']).group(1))
    #         self._full_disk_geolocation_metadata = geomet
    #     return self._full_disk_geolocation_metadata

    @property
    def geolocation_metadata(self):
        """Return geolocation metadata."""
        if not hasattr(self, "_geolocation_metadata"):
            geomet = {}
            geomet["ob_area"] = re.sub(
                r"\W+", "", self.metadata["block_2"]["projection"]
            )
            geomet["num_lines"] = self.metadata["block_1"]["num_lines"]
            geomet["num_samples"] = self.metadata["block_1"]["num_samples"]
            geomet["line_scale"] = self.metadata["block_2"]["line_scale"]
            geomet["sample_scale"] = self.metadata["block_2"]["sample_scale"]
            geomet["line_offset"] = self.metadata["block_2"]["line_offset"]
            geomet["sample_offset"] = self.metadata["block_2"]["sample_offset"]
            geomet["projection"] = str(self.metadata["block_2"]["projection"])
            geomet["start_datetime"] = self.start_datetime
            # Grabs the sub point longitude from the projection in the form
            #   proj(sublon)
            geomet["sublon"] = float(
                re.search(r"\((.+)\)", self.metadata["block_2"]["projection"]).group(1)
            )
            self._geolocation_metadata = geomet
        return self._geolocation_metadata

    @property
    def prologue(self):
        """Return prologue."""
        if not hasattr(self, "_prologue"):
            self._prologue = None
            if self.file_type == "prologue":
                self._prologue = self._read_prologue()
        return self._prologue

    # @property
    # def data(self):
    #     if not hasattr(self, '_data'):
    #         self._data = None
    #         if self.file_type == 'image':
    #             self._data = self._read_image_data()
    #     return self._data

    @property
    def epilogue(self):
        """Return epilogue."""
        if not hasattr(self, "_epilogue"):
            self._epilogue = None
            if self.file_type == "epilogue":
                self._epilogue = self._read_epilogue()
        return self._epilogue

    @property
    def name(self):
        """Return name."""
        return self._name

    @property
    def dirname(self):
        """Return file dirname."""
        return self._dirname

    @property
    def basename(self):
        """Return file basename."""
        return self._basename

    @property
    def file_type(self):
        """Return file_type."""
        return self._file_type_map[self.metadata["block_0"]["type_code"]]

    @property
    def start_datetime(self):
        """Return start_datetime."""
        return self.annotation_metadata["start_datetime"]

    @property
    def band(self):
        """Return band specified in block_128, if it exists."""
        if "block_128" in self.metadata:
            return "B{:02d}".format(self.metadata["block_128"]["chan_id"])
        else:
            return None

    @property
    def segment(self):
        """Return segment specified in block_128, if it exists."""
        if "block_128" in self.metadata:
            return self.metadata["block_128"]["segment"]
        else:
            return None

    @property
    def annotation_metadata(self):
        """Return annotation metadata (ie, platform, start time, etc)."""
        field_names = [
            "type",
            "disseminationID",
            "_",
            "platform",
            "band",
            "segment",
            "start_datetime",
            "compression",
        ]
        fields = {
            name: part.strip("_")
            for name, part in zip(
                field_names, self.metadata["block_4"]["annotation"].split("-")
            )
            if name != "_"
        }
        fields["platform"] = fields["platform"].lower()
        fields["start_datetime"] = datetime.strptime(
            fields["start_datetime"], "%Y%m%d%H%M"
        )
        return fields

    @property
    def _parts(self):
        """Return parts of filename, split on '-'."""
        return self.basename.split("-")

    @property
    def compressed(self):
        """Return True if compressed, False if not."""
        if "C" in self._parts[-1]:
            return True
        else:
            return False

    def decompress(self, outdir):
        """
        Decompress an xRIT file and return a file handle.

        The file will be decompressed to `outdir` and read from there.

        Returns an HritFile instance for the decompressed file.
        If already decompressed, raises an HritError.
        """
        # If compressed, then decompress
        if self.compressed:
            parts = copy(self._parts)
            parts[-1] = "__"
            new_fname = os.path.join(outdir, "-".join(parts))
            log.debug("Copying file to {}".format(new_fname))
            shutil.copy(self.name, new_fname)

            prevdir = os.getcwd()
            os.chdir(outdir)
            xRITDecompress(self.name)
            os.chdir(prevdir)

            return HritFile(new_fname)

        else:
            raise HritError("File already decompressed.  Unable to decompress.")

    def __read_metadata_block_info(self):
        """
        Read information about metadata blocks.

        This includes info about metadata blocks including which blocks are present
        and their lengths.  Return a dictionary whose keys are metadata block numbers
        and whose values are tuple containing the block's starting byte number and its
        length in bytes.
        """
        # Get the full header length (4 bytes starting at 4th byte)
        buff = self._fobj
        buff.seek(4)
        header_length = unpack(">I", buff.read(4))[0]

        # Start from beginning of file
        buff.seek(0)
        # First block is always 0th block and starts at first byte
        block_info = {0: (0, 0)}
        # Loop over blocks until we hit the end of the header
        first_loop = True
        while buff.tell() < header_length:
            # Read the block number and block length
            block_start = buff.tell()
            block_num = unpack("B", buff.read(1))[0]
            block_length = unpack(">H", buff.read(2))[0]

            if first_loop:
                if block_num != 0:
                    raise HritError(
                        "This does not appear to be an HRIT file. "
                        f"The first block number should be 0 but we got {block_num}."
                    )
                first_loop = False

            # Add block_length for the current block
            block_info[block_num] = (block_start, block_length)

            # Seek to the end of the block
            buff.seek(block_start + block_length)

        return block_info

    def __read_metadata(self):
        """Read the metadata from the file and return as a dictionary."""
        metadata = {}
        for block_num in self.block_info.keys():
            # Get the block map for the current block number
            # If it is not found, warn
            try:
                block_map = self.block_map[block_num]
                if block_map is None:
                    log.debug(
                        "Skipping known, but undefined metadata block: {}".format(
                            block_num
                        )
                    )
                    continue
            except KeyError:
                log.warning(
                    "Unrecognized metadata block number encountered: {}".format(
                        block_num
                    )
                )
                continue

            # Create metadata dictionary for block
            block_md = {}

            # Seek to the correct starting position in the buffer
            block_start, block_len = self.block_info[block_num]
            self._fobj.seek(block_start)

            # Populate metadata dictionary based on map
            for field_name, dtype, count in block_map:
                block_md[field_name] = np.fromfile(self._fobj, dtype=dtype, count=count)
                # Strip padding from strings
                if dtype.char == "S":
                    newlist = []
                    for valnum, val in enumerate(block_md[field_name]):
                        if field_name == "time":
                            newlist += [val.strip()]
                        else:
                            newlist += [val.decode("ascii").strip().replace("\x05", "")]
                    block_md[field_name] = newlist
                # If there is only one value, then return a scalar rather than an array
                if count == 1:
                    block_md[field_name] = block_md[field_name][0]

            # Add to metadata dictionary
            metadata["block_{}".format(block_num)] = block_md

        return metadata

    def _read_image_data(self, sector=None):
        """Read image data."""
        log.debug("Reading image file: {}".format(self.name))
        if self.file_type != "image":
            raise HritError(
                "Unable to read image data from file of type {}".format(self.file_type)
            )

        # Get to the start of the line record
        header_length = self.metadata["block_0"]["header_length"]
        self._fobj.seek(header_length)

        self._data = np.fliplr(np.array(list(read10bit(self._fobj))).reshape(464, 3712))
        return self._data

    def _read_prologue(self, sector=None):
        """Read prologue file."""
        log.debug("Reading prologue file: {}".format(self.name))
        if self.file_type != "prologue":
            raise HritError(
                "Unable to read prologue data from file of type {}".format(
                    self.file_type
                )
            )

        # Get to the start of the header record
        self._fobj.seek(self.metadata["block_0"]["header_length"])

        prologue = {}

        # SatelliteStatus block
        prologue["satelliteStatus"] = {}
        prologue["satelliteStatus"]["satelliteDefinition"] = {}
        satdef = prologue["satelliteStatus"]["satelliteDefinition"]
        satdef["satelliteID"] = self.__rf(">u2")
        satdef["nominalLongitude"] = self.__rf(">f4")
        satdef["satelliteStatus"] = self.__rf(">u1")

        prologue["satelliteStatus"]["satelliteOperations"] = {}
        satops = prologue["satelliteStatus"]["satelliteOperations"]
        satops["lastManeuverFlag"] = self.__rf(">u1")
        satops["lastManeuverStartTime"] = self.__read_time_cds()
        satops["lastManeuverEndTime"] = self.__read_time_cds()
        satops["lastManeuverType"] = self.__rf(">u1")
        satops["nextManeuverFlag"] = self.__rf(">u1")
        satops["nextManeuverStartTime"] = self.__read_time_cds()
        satops["nextManeuverEndTime"] = self.__read_time_cds()
        satops["nextManeuverType"] = self.__rf(">u1")

        prologue["satelliteStatus"]["orbit"] = {}
        orbit = prologue["satelliteStatus"]["orbit"]
        orbit["periodStartTime"] = self.__read_time_cds()
        orbit["periodEndTime"] = self.__read_time_cds()
        orbit["orbitPolynomial"] = self.__read_orbit_polynomial()

        prologue["satelliteStatus"]["attitude"] = {}
        attitude = prologue["satelliteStatus"]["attitude"]
        attitude["periodStartTime"] = self.__read_time_cds()
        attitude["periodEndTime"] = self.__read_time_cds()
        attitude["principleAxisOffsetAngle"] = self.__rf(">f8")
        attitude["attitudePolynomial"] = self.__read_attitude_polynomial()

        prologue["satelliteStatus"]["spinRateatRCStart"] = self.__rf(">f8")

        prologue["satelliteStatus"]["utcCorrelation"] = {}
        utcCorr = prologue["satelliteStatus"]["utcCorrelation"]
        utcCorr["periodStartTime"] = self.__read_time_cds()
        utcCorr["periodEndTime"] = self.__read_time_cds()
        utcCorr["onBoardTimeStart"] = self.__rf(">u1", shape=(7,))
        utcCorr["varOnBoardTimeStart"] = self.__rf(">f8")
        utcCorr["a1"] = self.__rf(">f8")
        utcCorr["varA1"] = self.__rf(">f8")
        utcCorr["a2"] = self.__rf(">f8")
        utcCorr["varA2"] = self.__rf(">f8")

        # ImageAcquisition block
        prologue["imageAcquisition"] = {}
        prologue["imageAcquisition"]["plannedAcquisitionTime"] = {}
        acqTime = prologue["imageAcquisition"]["plannedAcquisitionTime"]
        acqTime["trueRepeatCycleStart"] = self.__read_time_cds(True)
        acqTime["plannedForwardScanEnd"] = self.__read_time_cds(True)
        acqTime["plannedRepeatCycleEnd"] = self.__read_time_cds(True)

        prologue["imageAcquisition"]["radiometerStatus"] = {}
        radioStat = prologue["imageAcquisition"]["radiometerStatus"]
        radioStat["channelStatus"] = self.__rf(">u1", shape=(12,))
        radioStat["detectorStatus"] = self.__rf(">u1", shape=(42,))

        prologue["imageAcquisition"]["radiometerSettings"] = {}
        radioSett = prologue["imageAcquisition"]["radiometerSettings"]
        radioSett["MDUSamplingDelays"] = self.__rf(">u2", shape=(42,))

        radioSett["hrvFrameOffsets"] = {}
        frameOff = radioSett["hrvFrameOffsets"]
        frameOff["MDUNomHRVDelay1"] = self.__rf(">u2")
        frameOff["MDUNomHRVDelay2"] = self.__rf(">u2")
        frameOff["spare"] = self.__rf("|S2")
        frameOff["MDUNomHRVBreakline"] = self.__rf(">u2")
        radioSett["DHSSSyncSelection"] = self.__rf(">u1")
        radioSett["MDUOutGain"] = self.__rf(">u2", shape=(42,))
        radioSett["MDUCourseGain"] = self.__rf(">u1", shape=(42,))
        radioSett["MDUFinegain"] = self.__rf(">u2", shape=(42,))
        radioSett["MDUNumericalOffset"] = self.__rf(">u2", shape=(42,))
        radioSett["PUGain"] = self.__rf(">u2", shape=(42,))
        radioSett["PUOffset"] = self.__rf(">u2", shape=(27,))
        radioSett["PUBias"] = self.__rf(">u2", shape=(15,))

        radioSett["operationParameters"] = {}
        operParam = radioSett["operationParameters"]
        operParam["L0_LineCounter"] = self.__rf(">u2")
        operParam["K1_RetraceLines"] = self.__rf(">u2")
        operParam["K2_PauseDeciseconds"] = self.__rf(">u2")
        operParam["K3_RetraceLines"] = self.__rf(">u2")
        operParam["K4_PauseDeciseconds"] = self.__rf(">u2")
        operParam["K5_RetraceLines"] = self.__rf(">u2")
        operParam["X_DeepSpaceWindowPosition"] = self.__rf(">u1")

        radioSett["refocusingLines"] = self.__rf(">u2")
        radioSett["refocusingDirection"] = self.__rf(">u1")
        radioSett["refocusingPosition"] = self.__rf(">u2")
        radioSett["scanRefPosFlag"] = self.__rf(">u1")
        radioSett["scanRefPosNumber"] = self.__rf(">u2")
        radioSett["scanRefPotVal"] = self.__rf(">f4")
        radioSett["scanFirstLine"] = self.__rf(">u2")
        radioSett["scanLastLine"] = self.__rf(">u2")
        radioSett["retraceStartLine"] = self.__rf(">u2")

        prologue["imageAcquisition"]["radiometerOperations"] = {}
        radioOper = prologue["imageAcquisition"]["radiometerOperations"]
        radioOper["lastGainChangeFlag"] = self.__rf(">u1")
        radioOper["lastGainChangeTime"] = self.__read_time_cds()

        radioOper["decontamination"] = {}
        dcon = radioOper["decontamination"]
        dcon["decontaminationNow"] = self.__rf(">u1")
        dcon["decontaminationStart"] = self.__read_time_cds()
        dcon["decontaminationEnd"] = self.__read_time_cds()

        radioOper["bbCalScheduled"] = self.__rf(">u1")
        radioOper["bbCalibrationType"] = self.__rf(">u1")
        radioOper["bbFirstLine"] = self.__rf(">u2")
        radioOper["bbLastLine"] = self.__rf(">u2")
        radioOper["coldFocalPlaneOpTemp"] = self.__rf(">u2")
        radioOper["warmFocalPlaneOpTemp"] = self.__rf(">u2")

        # CelestialEvents block
        prologue["celestialEvents"] = {}
        prologue["celestialEvents"]["celestialBodiesPosition"] = {}
        celestBod = prologue["celestialEvents"]["celestialBodiesPosition"]
        celestBod["periodStartTime"] = self.__read_time_cds()
        celestBod["periodEndTime"] = self.__read_time_cds()
        celestBod["relatedOrbitFileTime"] = self.__rf("|S15")
        celestBod["relatedAttitudeFileTime"] = self.__rf("|S15")
        celestBod["earthEphemeris"] = self.__read_ephemeris()
        celestBod["moonEphemeris"] = self.__read_ephemeris()
        celestBod["sunEphemeris"] = self.__read_ephemeris()
        celestBod["starEphemeris"] = self.__read_starcoeff()

        prologue["celestialEvents"]["relationToImage"] = {}
        relToImg = prologue["celestialEvents"]["relationToImage"]
        relToImg["typeOfEclipse"] = self.__rf(">u1")
        relToImg["eclipseStartTime"] = self.__read_time_cds()
        relToImg["eclipseEndTime"] = self.__read_time_cds()
        relToImg["visibleBodiesInImage"] = self.__rf(">u1")
        relToImg["bodiesCloseToFOV"] = self.__rf(">u1")
        relToImg["impactOnImageQuality"] = self.__rf(">u1")

        # ImageDescription block
        prologue["imageDescription"] = {}
        prologue["imageDescription"]["projectionDescription"] = {}
        projDesc = prologue["imageDescription"]["projectionDescription"]
        projDesc["typeOfProjection"] = self.__rf(">u1")
        projDesc["longitudeOfSSP"] = self.__rf(">f4")

        prologue["imageDescription"]["referenceGridVIS_IR"] = {}
        refGrdVI = prologue["imageDescription"]["referenceGridVIS_IR"]
        refGrdVI["numberOfLines"] = self.__rf(">i4")
        refGrdVI["numberOfColumns"] = self.__rf(">i4")
        refGrdVI["lineDirGridStep"] = self.__rf(">f4")
        refGrdVI["columnDirGridStep"] = self.__rf(">f4")
        refGrdVI["gridOrigin"] = self.__rf(">u1")

        prologue["imageDescription"]["referenceGridHRV"] = {}
        refGrdHR = prologue["imageDescription"]["referenceGridHRV"]
        refGrdHR["numberOfLines"] = self.__rf(">i4")
        refGrdHR["numberOfColumns"] = self.__rf(">i4")
        refGrdHR["lineDirGridStep"] = self.__rf(">f4")
        refGrdHR["columnDirGridStep"] = self.__rf(">f4")
        refGrdHR["gridOrigin"] = self.__rf(">u1")

        prologue["imageDescription"]["plannedCoverageVIS_IR"] = {}
        pldCovVI = prologue["imageDescription"]["plannedCoverageVIS_IR"]
        pldCovVI["southernLinePlanned"] = self.__rf(">i4")
        pldCovVI["northernLinePlanned"] = self.__rf(">i4")
        pldCovVI["easternLinePlanned"] = self.__rf(">i4")
        pldCovVI["westernLinePlanned"] = self.__rf(">i4")

        prologue["imageDescription"]["plannedCoverageHRV"] = {}
        pldCovHR = prologue["imageDescription"]["plannedCoverageHRV"]
        pldCovHR["lowerSouthLinePlanned"] = self.__rf(">i4")
        pldCovHR["lowerNorthLinePlanned"] = self.__rf(">i4")
        pldCovHR["lowerEastLinePlanned"] = self.__rf(">i4")
        pldCovHR["lowerWestLinePlanned"] = self.__rf(">i4")
        pldCovHR["upperSouthLinePlanned"] = self.__rf(">i4")
        pldCovHR["upperNorthLinePlanned"] = self.__rf(">i4")
        pldCovHR["upperEastLinePlanned"] = self.__rf(">i4")
        pldCovHR["upperWestLinePlanned"] = self.__rf(">i4")

        prologue["imageDescription"]["level15ImageProduction"] = {}
        l15Proj = prologue["imageDescription"]["level15ImageProduction"]
        l15Proj["imageProcDirection"] = self.__rf(">u1")
        l15Proj["plannedGenDirection"] = self.__rf(">u1")
        l15Proj["plannedChanProcessing"] = self.__rf(">u1", shape=(12,))

        # RadiometricProcessing block
        prologue["radiometricProcessing"] = {}
        prologue["radiometricProcessing"]["rpSummary"] = {}
        rpSumm = prologue["radiometricProcessing"]["rpSummary"]
        rpSumm["radianceLinearization"] = self.__rf(">u1", shape=(12,))
        rpSumm["detectorEqualization"] = self.__rf(">u1", shape=(12,))
        rpSumm["onboardCalibrationResult"] = self.__rf(">u1", shape=(12,))
        rpSumm["mpefCalFeedback"] = self.__rf(">u1", shape=(12,))
        rpSumm["mtfAdatption"] = self.__rf(">u1", shape=(12,))
        rpSumm["straylightCorrectionFlag"] = self.__rf(">u1", shape=(12,))

        prologue["radiometricProcessing"][
            "level15ImageCalibration"
        ] = self.__read_image_calibration()

        prologue["radiometricProcessing"]["blackBodyDataUsed"] = {}
        bbUsed = prologue["radiometricProcessing"]["blackBodyDataUsed"]
        bbUsed["bbObservationUTC"] = self.__read_time_cds(True)

        bbUsed["bbRelatedData"] = {}
        bbRelDat = bbUsed["bbRelatedData"]
        bbRelDat["onBoardBBTime"] = self.__rf(">u1", shape=(7,))
        bbRelDat["mduOutGain"] = self.__rf(">u2", shape=(42,))
        bbRelDat["mduCoarseGain"] = self.__rf(">u1", shape=(42,))
        bbRelDat["mduFineGain"] = self.__rf(">u2", shape=(42,))
        bbRelDat["mduNumericalOffset"] = self.__rf(">u2", shape=(42,))
        bbRelDat["PUGain"] = self.__rf(">u2", shape=(42,))
        bbRelDat["PUOffset"] = self.__rf(">u2", shape=(27,))
        bbRelDat["PUBias"] = self.__rf(">u2", shape=(15,))
        bbRelDat["dcrValues"] = self.__rf("|S63")
        bbRelDat["xDeepSpaceWindowPosition"] = self.__rf(">u1")

        bbRelDat["coldFPTempterature"] = {}
        coldFPT = bbRelDat["coldFPTempterature"]
        coldFPT["fcuNominalColdFocalPlaneTemp"] = self.__rf(">u2")
        coldFPT["fcuRedundantColdFocalPlaneTemp"] = self.__rf(">u2")

        bbRelDat["warmFPTempterature"] = {}
        warmFPT = bbRelDat["warmFPTempterature"]
        warmFPT["fcuNominalWarmFocalPlaneVHROTemp"] = self.__rf(">u2")
        warmFPT["fcuRedundantWarmFocalPlaneVHROTemp"] = self.__rf(">u2")

        bbRelDat["scanMirrorTemperature"] = {}
        scanMT = bbRelDat["scanMirrorTemperature"]
        scanMT["fcuNominalScanMirrorSensor1Temp"] = self.__rf(">u2")
        scanMT["fcuRedundantScanMirrorSensor1Temp"] = self.__rf(">u2")
        scanMT["fcuNominalScanMirrorSensor2Temp"] = self.__rf(">u2")
        scanMT["fcuRedundantScanMirrorSensor2Temp"] = self.__rf(">u2")

        bbRelDat["m1m2m3Temperature"] = {}
        m1m2m3 = bbRelDat["m1m2m3Temperature"]
        m1m2m3["fcuNominalM1MirrorSensor1Temp"] = self.__rf(">u2")
        m1m2m3["fcuRedundantM1MirrorSensor1Temp"] = self.__rf(">u2")
        m1m2m3["fcuNominalM1MirrorSensor2Temp"] = self.__rf(">u2")
        m1m2m3["fcuRedundantM1MirrorSensor2Temp"] = self.__rf(">u2")
        m1m2m3["fcuNominalM23AssemblySensor1Temp"] = self.__rf(">u1")
        m1m2m3["fcuRedundantM23AssemblySensor1Temp"] = self.__rf(">u1")
        m1m2m3["fcuNominalM23AssemblySensor2Temp"] = self.__rf(">u1")
        m1m2m3["fcuRedundantM23AssemblySensor2Temp"] = self.__rf(">u1")

        bbRelDat["baffleTemperature"] = {}
        baffTemp = bbRelDat["baffleTemperature"]
        baffTemp["fcuNominalM1BaffleTemp"] = self.__rf(">u2")
        baffTemp["fcuRedundantM1BaffleTemp"] = self.__rf(">u2")

        bbRelDat["blackBodyTemperature"] = {}
        bbTemp = bbRelDat["blackBodyTemperature"]
        bbTemp["fcuNominalBlackBodySensorTemp"] = self.__rf(">u2")
        bbTemp["fcuRedundantBlackBodySensorTemp"] = self.__rf(">u2")

        bbRelDat["fcuMode"] = {}
        fcuMode = bbRelDat["fcuMode"]
        fcuMode["fcuNominalSMMStatus"] = self.__rf("|S2")
        fcuMode["fcuRedundantSMMStatus"] = self.__rf("|S2")

        bbRelDat["extractedBBData"] = self.__read_extracted_bb_data()

        prologue["radiometricProcessing"][
            "mpefCalFeedback"
        ] = self.__read_impf_cal_data()

        prologue["radiometricProcessing"]["radTransform"] = self.__rf(
            ">f4", shape=(42, 64)
        )

        prologue["radiometricProcessing"]["radProcMTFAdaption"] = {}
        radAdapt = prologue["radiometricProcessing"]["radProcMTFAdaption"]
        radAdapt["vis_irMTFCorrectionE_W"] = self.__rf(">f4", shape=(33, 16))
        radAdapt["vis_irMTFCorrectionN_S"] = self.__rf(">f4", shape=(33, 16))
        radAdapt["hrfMTFCorrectionE_W"] = self.__rf(">f4", shape=(9, 16))
        radAdapt["hrvMTFCorrectionN_S"] = self.__rf(">f4", shape=(9, 16))
        radAdapt["straylightCorrection"] = self.__rf(">f4", shape=(12, 8, 8))

        # GeometricProcessing block
        prologue["geometricProcessing"] = {}
        prologue["geometricProcessing"]["optAxisDistances"] = {}
        optAxDist = prologue["geometricProcessing"]["optAxisDistances"]
        optAxDist["e-wFocalPlane"] = self.__rf(">f4", shape=(42,))
        optAxDist["n-sFocalPlane"] = self.__rf(">f4", shape=(42,))

        prologue["geometricProcessing"]["earthModel"] = {}
        earthMod = prologue["geometricProcessing"]["earthModel"]
        earthMod["typeOfEarthModel"] = self.__rf(">u1")
        earthMod["equatorialRadius"] = self.__rf(">f8")
        earthMod["northPolarRadius"] = self.__rf(">f8")
        earthMod["southPolarRadius"] = self.__rf(">f8")

        prologue["geometricProcessing"]["atmosphericModel"] = self.__rf(
            ">f4", shape=(12, 360)
        )
        prologue["geometricProcessing"]["resamplingFunctions"] = self.__rf(
            ">u1", shape=(12,)
        )

        self._prologue = prologue
        return self._prologue

    def _read_epilogue(self, sector=None):
        """Read epilogue file."""
        log.debug("Reading epilogue file: {}".format(self.name))
        if self.file_type != "epilogue":
            raise HritError(
                "Unable to read epilogue data from file of type {}".format(
                    self.file_type
                )
            )

        return None

    # Routine for reading specific things
    def __read_field(self, dtype, shape=(1,)):
        """Read field."""
        dtype = np.dtype(dtype)
        bytes_per_elem = dtype.itemsize
        nbytes = bytes_per_elem * reduce(operator.mul, shape, 1)

        dat = np.frombuffer(self._fobj.read(nbytes), dtype=dtype).reshape(shape)
        if dat.size == 1:
            dat = dat[0]

        return dat

    __rf = __read_field

    def __read_time_cds(self, expanded=False):
        """
        Read time CDS.

        Each self.__rf must remain, even if unused, since the reads increment
        the pointer.
        """
        epoch = datetime(1958, 1, 1, 0, 0, 0)
        days = self.__rf(">u2")
        millisec = self.__rf(">u4")
        # Must read in nanosec, even though it is never used.
        # Otherwise data will be out of sync from missing bytes in
        # the binary read.
        if expanded:
            microsec = self.__rf(">u2")
            nanosec = self.__rf(">u2")  # NOQA
        else:
            microsec = 0
            nanosec = 0  # NOQA
        time = epoch + timedelta(
            days=int(days), milliseconds=int(millisec), microseconds=int(microsec)
        )
        return time

    def __read_orbit_polynomial(self):
        """Read orbit polynomial."""
        poly = []
        for elemind in range(0, 100):
            elem = {
                "startTime": self.__read_time_cds(),
                "endTime": self.__read_time_cds(),
                "X": self.__rf(">f8", shape=(8,)),
                "Y": self.__rf(">f8", shape=(8,)),
                "Z": self.__rf(">f8", shape=(8,)),
                "VX": self.__rf(">f8", shape=(8,)),
                "VY": self.__rf(">f8", shape=(8,)),
                "VZ": self.__rf(">f8", shape=(8,)),
            }
            poly.append(elem)
        return poly

    def __read_attitude_polynomial(self):
        """Read attitude polynomial."""
        poly = []
        for elemind in range(0, 100):
            elem = {
                "startTime": self.__read_time_cds(),
                "endTime": self.__read_time_cds(),
                "XofSpinAxis": self.__rf(">f8", shape=(8,)),
                "YofSpinAxis": self.__rf(">f8", shape=(8,)),
                "ZofSpinAxis": self.__rf(">f8", shape=(8,)),
            }
            poly.append(elem)
        return poly

    def __read_ephemeris(self):
        """Read ephemeris."""
        ephem = []
        for elemind in range(0, 100):
            elem = {
                "startTime": self.__read_time_cds(),
                "endTime": self.__read_time_cds(),
                "alphaCoef": self.__rf(">f8", shape=(8,)),
                "betaCoef": self.__rf(">f8", shape=(8,)),
            }
            ephem.append(elem)
        return ephem

    def __read_starcoeff(self):
        """Read star coefficient."""
        coeff = []
        for coeffind in range(0, 100):
            stars = []
            for starind in range(0, 20):
                elem = {
                    "starID": self.__rf(">u2"),
                    "startTime": self.__read_time_cds(),
                    "endTime": self.__read_time_cds(),
                    "alphaCoef": self.__rf(">f8", shape=(8,)),
                    "betaCoef": self.__rf(">f8", shape=(8,)),
                }
                stars.append(elem)
            coeff.append(stars)
        return coeff

    def __read_image_calibration(self):
        """Read image calibration."""
        cal = []
        for chind in range(0, 12):
            chcal = {}
            chcal["slope"] = self.__rf(">f8")
            chcal["offset"] = self.__rf(">f8")
            cal.append(chcal)
        return cal

    def __read_extracted_bb_data(self):
        """Read extracted BB data."""
        bb_data = []
        for chind in range(0, 12):
            dat = {}
            dat["numberOfPixelsUsed"] = self.__rf(">u4")
            dat["meanCount"] = self.__rf(">f4")
            dat["rms"] = self.__rf(">f4")
            dat["maxCount"] = self.__rf(">u2")
            dat["minCount"] = self.__rf(">u2")
            dat["bbProcessingSlope"] = self.__rf(">f8")
            dat["bbProcessingOffset"] = self.__rf(">f8")
            bb_data.append(dat)
        return bb_data

    def __read_impf_cal_data(self):
        """Read IMPF Calibration data."""
        cal_data = []
        for chind in range(0, 12):
            cal = {}
            cal["imageQualityFlag"] = self.__rf(">u1")
            cal["referenceDataFlag"] = self.__rf(">u1")
            cal["absCalMethod"] = self.__rf(">u1")
            cal["pad1"] = self.__rf("|S1")
            cal["absCalWeightVic"] = self.__rf(">f4")
            cal["absCalWeightXsat"] = self.__rf(">f4")
            cal["absCalCoeff"] = self.__rf(">f4")
            cal["absCalError"] = self.__rf(">f4")
            cal["gscisCalCoeff"] = self.__rf(">f4")
            cal["gscisCalError"] = self.__rf(">f4")
            cal["gscisOffsetCount"] = self.__rf(">f4")
            cal_data.append(cal)
        return cal_data



#!/bin/env python

# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Update release note index with latest version release reference."""
from sys import argv
from glob import glob
from pathlib import Path
from os.path import basename, exists


def main(index_filename, release_note_path):
    """Generate an index.rst file for release notes.

    This function scans the specified directory for release note files, sorts them based
    on version numbers (and alpha releases), and generates an
    ``index.rst`` file that includes a table of contents.

    Parameters
    ----------
    index_filename : str or Path
        The path to the output ``index.rst`` file to be generated.
    release_note_path : str or Path
        The directory containing release note ``.rst`` files.

    Notes
    -----
    The function processes release note files that match the following naming patterns:

    - ``vX_Y_Z.rst``
    - ``vX_Y_ZaN.rst``
    - ``X.Y.Z.rst``
    - ``X.Y.ZaN.rst``

    where ``X``, ``Y``, ``Z``, and ``N`` are integers representing the major, minor,
    bugfix, and alpha version numbers, respectively.

    Release notes not matching these patterns are ignored.

    If a file named ``latest.rst`` exists in the ``release_note_path``, it is included
    at the top of the generated index under the heading "Latest (version on cutting edge
    of git)".

    It excludes ``index.rst`` and ``latest.rst`` from the list of release notes
    to be processed (except for including ``latest.rst`` at the top if it exists).

    The release notes are sorted in reverse chronological order (newest first) based on
    their version numbers.

    The function does not include high-level version summaries from
    ``version_summaries.yaml``; these are not currently populated in ``index.rst``.
    These will be included when brassy takes over this functionality.

    Examples
    --------
    Generate an ``index.rst`` file from release notes in ``docs/releases``:

    >>> from pathlib import Path
    >>> index_file = 'docs/releases/index.rst'
    >>> release_notes_dir = 'docs/releases'
    >>> main(index_file, release_notes_dir)
    """
    index_filename = Path(index_filename)
    release_note_path = Path(release_note_path)
    # Note this does NOT include the high level version summaries - those live
    # in docs/source/releases/version_summaries.yaml, but are not currently being
    # populated in index.rst.  When this release note index creation is moved to
    # brassy, the version summaries will be pulled from version_summaries.yaml
    # to include in the index.rst.

    # Note this will sort release notes properly named as:
    # * vX_Y_Z.rst
    # * vX_Y_ZaN.rst
    # * X.Y.Z.rst
    # * X.Y.ZaN.rst

    # Includes "latest" at the top if it exists.

    # No other release note name formatting supported.

    # Include standard top level headers
    lines = [".. dropdown:: Distribution Statement\n"]
    lines += ["\n"]
    lines += [" | This is an auto-generated file.\n"]
    lines += [" | Please abide by license packaged with this software.\n"]
    lines += ["\n"]
    lines += [".. _release_notes:\n"]
    lines += ["\n"]
    lines += ["Release Notes\n"]
    lines += ["*************\n"]
    lines += ["\n"]

    # If latest.rst exists, include it in the index.rst
    if exists(str(release_note_path / "latest.rst")):
        lines += ["Latest (version on cutting edge of git)\n"]
        lines += ["---------------------------------------\n"]
        lines += ["\n"]
        lines += [".. toctree::\n"]
        lines += ["   :maxdepth: 1\n"]
        lines += ["\n"]
        lines += ["   latest\n"]
        lines += ["\n"]

    # If upcoming.rst exists, include it in the index.rst
    if exists(str(release_note_path / "upcoming.rst")):
        lines += ["Upcoming (version AHEAD OF the cutting edge of git)\n"]
        lines += ["---------------------------------------------------\n"]
        lines += ["\n"]
        lines += [".. toctree::\n"]
        lines += ["   :maxdepth: 1\n"]
        lines += ["\n"]
        lines += ["   upcoming\n"]
        lines += ["\n"]

    # Collect all the major, minor, bugfix, alpha version numbers, in a sortable
    # manner
    versions = {}
    # List all the release notes of format v*.rst - ie, don't include index.rst
    # or latest.rst, but DO include X_Y_ZaN.rst (alpha releases)
    for rst_file in glob(str(release_note_path / "*.rst")):
        if basename(rst_file) in ["latest.rst", "index.rst", "upcoming.rst"]:
            continue
        # print(f"Adding {rst_file}")
        # Just leave X.Y.ZaN or X.Y.Z, as appropriate
        # This will work with vX_Y_Z.rst, vX_Y_ZaN.rst, X.Y.Z.rst, or X.Y.ZaN.rst
        version_num = (
            basename(rst_file).replace("v", "").replace(".rst", "").replace("_", ".")
        )
        # print(f"verson {version_num}")
        # Default alpha to None and mmb (major/minor/bugfix) to version_num -
        # alpha not defined for full releases.
        alpha = None
        mmb_version_num = version_num
        # If there is an "a" in version_num, ie X_Y_ZaN, ensure we capture the
        # appropriate alpha version, and remove the aN from the mmb_version_num
        if "a" in version_num:
            mmb_version_num, alpha = version_num.split("a")
        # Now split for major/minor/bugfix from the X_Y_Z version
        major, minor, bugfix = mmb_version_num.split(".")
        # Zero pad these for sorting
        major = f"{int(major):03}"
        minor = f"{int(minor):03}"
        bugfix = f"{int(bugfix):03}"

        # Only include alpha if it is defined.
        if alpha is not None:
            alpha = f"{int(alpha):03}"
        # Store these in dictionaries so we can loop through in order
        if major not in versions:
            versions[major] = {}
        if minor not in versions[major]:
            versions[major][minor] = {}
        if bugfix not in versions[major][minor]:
            versions[major][minor][bugfix] = []
        # If alphas are defined, store them in a list at the lowest level.
        if alpha is not None:
            versions[major][minor][bugfix] += [alpha]

    # Now loop through major, minor, bugfix, and alpha, appending sections
    # in order as appropriate.
    for major in sorted(versions, reverse=True):
        for minor in sorted(versions[major], reverse=True):
            print(f"Adding {int(major)}.{int(minor)}")
            version_header = f"Version {int(major)}.{int(minor)}"
            version_underline = "-" * (len(version_header))
            lines += [f"{version_header}\n"]
            lines += [f"{version_underline}\n"]
            lines += ["\n"]
            lines += [".. toctree::\n"]
            lines += ["   :maxdepth: 1\n"]
            lines += ["\n"]
            for bugfix in sorted(versions[major][minor], reverse=True):
                curr_mmb = f"v{int(major)}_{int(minor)}_{int(bugfix)}"
                curr_mmb_period = f"{int(major)}.{int(minor)}.{int(bugfix)}"
                # Only add major/minor/bugfix version if it exists - could be
                # an alpha release with NO MMB release.
                if exists(str(release_note_path / f"{curr_mmb}.rst")):
                    lines += [f"   {curr_mmb}\n"]
                elif exists(str(release_note_path / f"{curr_mmb_period}.rst")):
                    lines += [f"   {curr_mmb_period}\n"]
                for alpha in sorted(versions[major][minor][bugfix], reverse=True):
                    if exists(str(release_note_path / f"{curr_mmb}a{int(alpha)}.rst")):
                        lines += [f"   {curr_mmb}a{int(alpha)}\n"]
                    elif exists(
                        str(release_note_path / f"{curr_mmb_period}a{int(alpha)}.rst")
                    ):
                        lines += [f"   {curr_mmb_period}a{int(alpha)}\n"]
            lines += ["\n"]
    # Now write all the lines we've collected out to index.rst
    with open(index_filename, "w") as fobj:
        print(f"Writing {index_filename}")
        fobj.writelines(lines)


if __name__ == "__main__":
    main(argv[1], argv[2])



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Reader to read a grannual NOAA ATMS SDR TBs in h5 format.

Output variables in xarray object for geoips processing system

V0:   August 25, 2021

The date is generated by the NOAA community satellite processing package
(CSPP), developed at CIMSS

Example of ATMS file names::

    'SATMS_j01_d20210809_t0959306_e1000023_b19296_fnmoc_ops.h5'
        SDR TBs variables
    'GATMO_j01_d20210809_t0959306_e1000023_b19296_fnmoc_ops.h5'
        SDR Geolocation variables

Dataset info::

    TB[12,96,22]:  for each granuel

    CHAN#  Center-Freq(GHz)  POL
    1      23.8               V
    2      31.4               V
    3      50.3               H
    4      51.76              H
    5      52.8               H
    6      53.596+-0.115      H
    7      54.4               H
    8      54.94              H
    9      55.5               H
    10     57.290(f0)         H
    11     f0 +-0.217         H
    12     f0 +-0.322+-0.048  H
    13     f0 +-0.322+-0.022  H
    14     f0 +-0.322+-0.010  H
    15     f0 +-0.322+-0.0045 H

    16     88.2               V
    17     165.5              H
    18     183.1+-7           H
    19     183.1+-4.5         H    (FNMOC used this chan for 183 GHz image)
    20     183.1+-3.0         H
    21     183.1+-1.8         H
    22     183.1+-1.0         H

    BeamTime[12,96]:  microsecond, i.e., 1*10^-6.  IET "IDPS Epoch Time" is used
                      It is a signed 64-bit integer giving microseconds since
                      00:00:00.000000 Jan 1 1958.
    BrightnessTemperatureFactors[2]: 1: scale (unitless); 2: offset (K)
    BrightnessTemperature[12,96,22]: [scan,pix,chans]

    SDR geolocation Info
    Latitude/Longitude[12,96]:   for Chan 17 only (for initial product,
                                                   it is used for all channels)
    BeamLatitude[12,96,5]:       for chan 1,2,3,16,17. They will be used for
                                            associated TBs at a later date.
    BeamLongitude[12,96,5]:      for chan 1,2,3,16,17
    SatelliteAzimuthAngle, SatelliteZenithAngle,
    SolarAzimuthAngle, SolarZenithAngle[12,96]

Notes
-----
Unix epoch time is defined as the number of seconds that have elapsed
since January 1, 1970 (midnight UTC/GMT). Thus, there is a 12 years
difference for the JPSS data when
datetime.datetime.utcfromtimestamp(epoch) is used
to convert the JPSS IDPS Epoch time to the humman-readable date.

This reader is developed to read one granual a time from ATMS npp and
jpss-1(n20) data files.

The example files are:

* ``SATMS_j01_d20210809_t0959306_e1000023_b19296_fnmoc_ops.h5``: for TBs.  'b': orbit#
* ``GATMO_j01_d20210809_t0959306_e1000023_b19296_fnmoc_ops.h5``: for geolocations
"""

# Python Standard Libraries
import datetime
import logging
from os.path import basename

# Third-Party Libraries
from astropy.time import Time
from dateutil.relativedelta import relativedelta
import h5py
import numpy as np
import xarray as xr

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "atms_hdf5"
source_names = ["atms"]

# from IPython import embed as shell

# list of variables selected from input files
atms_vars = [
    "BrightnessTemperature",
    "BrightnessTemperatureFactors",
    "BeamTime",
    "Latitude",
    "Longitude",
    "SatelliteAzimuthAngle",
    "SatelliteZenithAngle",
    "SolarAzimuthAngle",
    "SolarZenithAngle",
    "StartTime",
]

# unify common names for sat/sun-zenith and azimuth angles
xvarnames = {
    "SolarZenithAngle": "solar_zenith_angle",
    "SolarAzimuthAngle": "solar_azimuth_angle",
    "SatelliteZenithAngle": "satellite_zenith_angle",
    "SatelliteAzimuthAngle": "satellite_azimuth_angle",
}

# final_xarray = xr.Dataset()  # define a xarray to hold all selected variables


def convert_epoch_to_datetime64(time_array, use_shape=None):
    """Convert time to datetime object.

    Parameters
    ----------
    time_array : array
        Array of start time integers (multiplied by 1e-6 in function)
    use_shape : tuple, optional
        desired output shape of time array, by default None

    Returns
    -------
    array
        array of converted datetime objects
    """
    # Convert TAI to UTC using astropy
    utc_time = Time(time_array / 1e6, format="unix_tai").utc.datetime

    # ATMS Epoch starts at 1958-01-01, not 1970-01-01
    utc_time -= relativedelta(years=12)

    # Either convert 1D array to 2D, or return original shape
    if time_array.ndim == 1 and use_shape:
        nscan, npix = use_shape
        converted_time = np.zeros((nscan, npix)).astype(datetime.datetime)
        for i in range(nscan):
            converted_time[i, :] = utc_time[i]
    else:
        converted_time = utc_time

    return converted_time.astype(np.datetime64)


def read_atms_file(fname, xarray_atms, metadata_only=False):
    """Read ATCF data from file fname."""
    fileobj = h5py.File(fname, mode="r")

    # check for available variables from nput file
    # Do not read the BT data unless metadata_only is False
    if "ATMS-SDR_All" in fileobj["All_Data"].keys():  # for TB-data
        data_select = fileobj["All_Data"]["ATMS-SDR_All"]

        tb_time = data_select["BeamTime"][()]
        #  get UTC time in datetime64 format required by geoips for each pixel
        time_scan = convert_epoch_to_datetime64(tb_time)

        # Rather than reading from the file, create empty arrays of time_scan shape
        if metadata_only:
            V23 = V31 = H50 = V89 = H165 = H183 = np.empty(time_scan.shape)
        else:
            tb = data_select["BrightnessTemperature"][()]
            tb_factor = data_select["BrightnessTemperatureFactors"][()]

            # convert tb to actual values
            tbs = tb * tb_factor[0] + tb_factor[1]

            # TBs for selected channels
            V23 = tbs[:, :, 0]
            V31 = tbs[:, :, 1]
            H50 = tbs[:, :, 2]
            V89 = tbs[:, :, 15]
            H165 = tbs[:, :, 16]
            H183 = tbs[:, :, 18]  # to match the 183+-4.5 GHz channel used by FNMOC

        # make dict of numpy arrays
        ingested = {
            "V23": V23,
            "V31": V31,
            "H50": H50,
            "V89": V89,
            "H165": H165,
            "H183": H183,
            "sdr_time": time_scan,
        }

    if "ATMS-SDR-GEO_All" in fileobj["All_Data"].keys():  # for geo-data
        data_select = fileobj["All_Data"]["ATMS-SDR-GEO_All"]

        # We will always read in latitude regardless if metadata_only is True.
        # Do this so we can filter out bad data, which can cause the start time
        # to be somewhere around 1957....
        lat = data_select["Latitude"][()]
        StartTime = data_select["StartTime"][()]
        #  get UTC time in datetime64 format required by geoips for each pixel
        time_scan = convert_epoch_to_datetime64(StartTime, use_shape=lat.shape)

        ingested = {
            "latitude": lat,
            "geo_time": time_scan,
            "valid_indices": ~(abs(lat) > 90),
        }
        if not metadata_only:
            ingested["longitude"] = data_select["Longitude"][()]
            ingested["solar_zenith_angle"] = data_select["SolarZenithAngle"][()]
            ingested["solar_azimuth_angle"] = data_select["SolarAzimuthAngle"][()]
            ingested["satellite_zenith_angle"] = data_select["SatelliteZenithAngle"][()]
            ingested["satellite_azimuth_angle"] = data_select["SatelliteAzimuthAngle"][
                ()
            ]

    platform_name = fileobj.attrs["Platform_Short_Name"][0, 0].decode("utf-8")

    # close the h5 object
    fileobj.close()
    # final_xarray = xr.Dataset()
    final_xarray = xarray_atms
    #          ------  setup xarray variables   ------
    for key, data in ingested.items():
        if key not in xarray_atms.variables.keys():
            final_xarray[key] = xr.DataArray(data)
        else:
            merged_array = np.vstack([xarray_atms[key], data])
            final_xarray[key] = xr.DataArray(
                merged_array, dims=["dim_" + str(merged_array.shape[0]), "dim_1"]
            )
    # Set official time to either sdr_time or geo_time
    final_keys = final_xarray.variables.keys()
    if "sdr_time" in final_keys:
        final_xarray["time"] = final_xarray["sdr_time"]
    else:
        final_xarray["time"] = final_xarray["geo_time"]

    # Map platform names found in data file attributes to consistent
    # GeoIPS naming conventions (match VIIRS reader).
    if platform_name == "J01":
        platform_name = "noaa-20"
    elif platform_name == "J02":
        platform_name = "noaa-21"
    elif platform_name == "NPP":
        platform_name = "npp"
    final_xarray.attrs["platform_name"] = platform_name
    final_xarray.attrs["data_provider"] = "NOAA"

    return final_xarray


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read ATMS hdf5 data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    LOG.info("Reading files %s", fnames)

    # if metadata_only is True:  # read-in datafiles first time if 'metadata_only= True'
    xarray_atms = xr.Dataset()
    source_file_names = []
    for fname in fnames:
        xarray_atms = read_atms_file(fname, xarray_atms, metadata_only=metadata_only)
        source_file_names += [basename(fname)]  # name of last file from input files
    xarray_atms.attrs["source_file_names"] = source_file_names

    # setup attributors
    from geoips.xarray_utils.time import (
        get_max_from_xarray_time,
        get_min_from_xarray_time,
    )

    if "valid_indices" in xarray_atms:
        xarray_atms = xarray_atms.where(xarray_atms["valid_indices"])

    xarray_atms.attrs["start_datetime"] = get_min_from_xarray_time(xarray_atms, "time")
    xarray_atms.attrs["end_datetime"] = get_max_from_xarray_time(xarray_atms, "time")
    xarray_atms.attrs["source_name"] = "atms"

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for
    # final image
    xarray_atms.attrs["sample_distance_km"] = 2
    xarray_atms.attrs["interpolation_radius_of_influence"] = (
        30000  # could be tuned if needed
    )

    return {"METADATA": xarray_atms[[]], "ATMS": xarray_atms}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS CLI "test" command.

Runs the appropriate tests based on the arguments provided.
"""

from glob import glob
from importlib import resources

# from os import listdir
from os import environ, makedirs
from os.path import basename, exists, join
import sys

# from pytest import main as invoke_pytest
from subprocess import call

from geoips.commandline.geoips_command import GeoipsCommand, GeoipsExecutableCommand
from geoips.errors import PluginError
from geoips.geoips_utils import is_editable
from geoips.interfaces import sectors


# class GeoipsTestUnitTest(GeoipsExecutableCommand):
#     """Test Command for running GeoIPS Unit Tests."""

#     name = "unit-test"
#     command_classes = []

#     def add_arguments(self):
#         """Instantiate the arguments that are supported for the test unit-test command. # NOQA

#         Currently the "geoips test unit-test" command supports this format:
#             - geoips test unit-test dir_name <-p> <package_name> <-n> <test_name>
#         Where:
#             - dir_name is the name of the folder containing the unit-test[s] you want to # NOQA
#               run
#             - <package_name> is any GeoIPS package that is installed and recognized by
#               the GeoIPS Library
#             - <test_name> is the name of the unit test being ran
#         """
#         self.parser.add_argument(
#             "directory_name",
#             type=str,
#             help="GeoIPS Packages Unit Test Directory Name where unit tests are held.", # NOQA
#         )
#         self.parser.add_argument(
#             "--package_name",
#             "-p",
#             type=str,
#             default="geoips",
#             choices=self.plugin_package_names,
#             help="GeoIPS Package containing the unit-tests to be ran.",
#         )
#         self.parser.add_argument(
#             "--name_of_test",
#             "-n",
#             type=str,
#             default=None,
#             help="Specific Unit Test to be ran out of directory_name.",
#         )

#     def __call__(self, args):
#         """Run the provided unit tests based on the arguments provided."""
#         dir_name = args.directory_name
#         package_name = args.package_name
#         test_name = args.name_of_test
#         unit_test_dir = str(resources.files(package_name) / "../tests/unit_tests")

#         try:
#             # Try listing the expected unit test directory. If it fails, raise an
#             # argparse error which states such package doesn't have a unit tests
#             # directory
#             listdir(unit_test_dir)
#         except FileNotFoundError:
#             err_str = f"No unit tests directory found for package '{package_name}'."
#             self.parser.error(err_str)

#         if dir_name not in listdir(unit_test_dir):
#             # The specified unit test directory does not exist at the specified location # NOQA
#             # raise an error specifying that
#             err_str = f"Directory '{dir_name}' not found under {package_name}'s unit "
#             err_str += f"tests directory '{unit_test_dir}'. Please select one of the "
#             err_str += f"following unit test directories:\n {listdir(unit_test_dir)}"
#             self.parser.error(err_str)
#         elif test_name is not None:
#             # We've specified a specific Unit Test to run out of
#             # <package_name>/tests/unit_tests/<dir_name>/<script_name>, ensure that
#             # file actually exists
#             fnames = [
#                 basename(fpath)
#                 for fpath in glob(f"{unit_test_dir}/{dir_name}/test_*.py")
#             ]
#             if test_name not in fnames:
#                 err_str = f"Unit Test '{test_name}' not found under the directory "
#                 err_str += f"'{unit_test_dir}', please select one of the options shown " # NOQA
#                 err_str += f"below.\ns {fnames}"
#                 self.parser.error(err_str)
#         else:
#             # script name wasn't specified, run all unit tests found under
#             # <package_name>/tests/unit_tests/<dir_name>
#             test_name = "."

#         test_path = str(f"{unit_test_dir}/{dir_name}/{test_name}")
#         invoke_pytest(["-v", test_path])


class GeoipsTestSector(GeoipsExecutableCommand):
    """Test Command for creating a sector image based on the provided sector name.

    This used to be ran via 'create_sector_image', however we are trying to consolidate
    all independent console scripts to be used via the CLI. When this command is called
    an image of the provided sector will be created so we can view whether or not it
    matches the region of the globe we'd like to study.
    """

    name = "sector"
    command_classes = []

    def add_arguments(self):
        """Instantiate the arguments that are supported for the test sector command.

        Currently the "geoips test sector" command supports this format:
            - geoips test sector <sector_name> --outdir <output_directory_path>
        Where:
            - <sector_name> is the name of any GeoIPS Sector Plugin that has an entry in
              any package's plugin registry.
            - --outdir is the full path to the directory in which you'd like to create
              the sector image.
        """
        self.parser.add_argument(
            "sector_name",
            type=str,
            help="Name of the sector plugin to create an image from.",
        )
        self.parser.add_argument(
            "--outdir",
            "-o",
            type=str,
            default=f"{environ['GEOIPS_OUTDIRS']}",
            help="The output directory to create your sector image in.",
        )

    def __call__(self, args):
        """Create the provided sector image based off the arguments provided.

        This will retrieve the selected sector plugin from any GeoIPS Plugin package,
        then create an image of that sector. This is a good way to quickly test whether
        or not your sector plugin covers the area you expected with the correct
        resolution.

        Parameters
        ----------
        args: Argparse Namespace()
            - The list argument namespace to parse through
        """
        sector_name = args.sector_name
        outdir = args.outdir
        # If the path to outdir doesn't already exist, make that path
        if not exists(outdir):
            makedirs(outdir)
        # Create an image for the requested sector, including just the map and white
        # background.
        fname = join(outdir, f"{sector_name}.png")
        try:
            if "non_existent" in sector_name:
                # This occurs for a unit test that we are just checking the error output
                # for. No need to rebuild the plugin registry, which can be specified by
                # using rebuild_registries=False
                rebuild_registries = False
            else:
                # Otherwise, assume this is a new sector that is being developed, and
                # automate plugin registry creation if it does not already exist as an
                # entry in the registry.
                rebuild_registries = True
            sect = sectors.get_plugin(
                sector_name, rebuild_registries=rebuild_registries
            )
        except PluginError:
            raise self.parser.error(
                f"Sector '{sector_name}' is not a valid plugin.\nPlease use a plugin "
                "found under 'geoips list interface sectors' or create a new plugin "
                f"named '{sector_name}' and run 'create_plugin_registries'."
            )
        print(f"Creating {fname}.")
        sect.create_test_plot(fname)


class GeoipsTestScript(GeoipsExecutableCommand):
    """Test Command for running GeoIPS Test Scripts."""

    name = "script"
    command_classes = []

    def add_arguments(self):
        """Instantiate the arguments that are supported for the test script command.

        Currently the "geoips test script" command supports this format:
            - geoips test script <-p> <package_name> <script_name> <--integration>
        Where:
            - <package_name> is any GeoIPS package that is installed and recognized by
              the GeoIPS Library
            - <script_name> is the name of the bash script being tested
            - '--integration' represents whether or not this is an 'integration' test
        """
        self.parser.add_argument(
            "script_name",
            type=str,
            help="GeoIPS Script to be tested",
        )
        self.parser.add_argument(
            "--package_name",
            "-p",
            type=str,
            default="geoips",
            choices=self.plugin_package_names,
            help="GeoIPS Package containing the script to be tested",
        )
        self.parser.add_argument(
            "--integration",
            default=False,
            action="store_true",
            help="Whether or not the script is an integration test.",
        )

    def __call__(self, args):
        """Run the provided test script based off the arguments provided.

        This will run the provided test script, locating such script based on whether
        or not this was an integration-based test. If it wasn't, we'll look through the
        corresponding <package_name>/tests/scripts/ directory for the correct test
        script.

        Parameters
        ----------
        args: Argparse Namespace()
            - The list argument namespace to parse through
        """
        package_name = args.package_name
        script_name = args.script_name
        is_integration_test = args.integration

        if is_integration_test:
            if package_name != "geoips":
                # Raise an argparse error which states only geoips supports integration
                # testing
                err_str = "Only package 'geoips' has integration tests. Package "
                err_str += f"'{package_name}' doesn't have those tests. Try again."
                self.parser.error(err_str)
            dir_name = "integration_tests"
        else:
            dir_name = "scripts"
        if not is_editable(package_name):
            # Package is installed in non-editable mode and we will not be able to
            # access unit tests. Raise a runtime error reporting this.
            print(
                f"Error: Package '{package_name}' is installed in non-editable mode and"
                " we are not able to access it's unit tests. For this command to "
                f"work, please install '{package_name}' in editable mode via: "
                f"'pip install -e <path_to_{package_name}>'",
                file=sys.stderr,
            )
            # We use a print to sys.stderr so monkeypatch unit tests can catch this
            # output
            raise RuntimeError(
                f"Package '{package_name}' isn't installed in editable mode."
            )
        test_dir = str(resources.files(package_name) / f"../tests/{dir_name}")
        fnames = [basename(fpath) for fpath in glob(f"{test_dir}/*.sh")]

        if script_name not in fnames:
            # Raise an argparse error which states that file doesn't exist within the
            # specified directory
            str_fnames = ",\n".join(fnames)
            err_str = f"Script '{script_name}' doesn't exist within '{test_dir}/'. "
            err_str += "Please select a valid script from this directory listing:\n"
            err_str += str_fnames
            self.parser.error(err_str)

        script_path = f"{test_dir}/{script_name}"
        # Run the corresponding test
        call(["bash", script_path], shell=False)


class GeoipsTestLinting(GeoipsExecutableCommand):
    """Test Command for running GeoIPS Linting Services."""

    name = "linting"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the test-subparser for the Test Linting Command."""
        self.parser.add_argument(
            "--package_name",
            "-p",
            type=str,
            default="geoips",
            choices=self.plugin_package_names,
            help="GeoIPS Package that we want to run linting tests on.",
        )

    def __call__(self, args):
        """Run all GeoIPS Linting Tests on the provided package."""
        package_name = args.package_name
        if not is_editable(package_name):
            # Package is installed in non-editable mode and we will not be able to
            # access unit tests. Raise a runtime error reporting this.
            print(
                f"Error: Package '{package_name}' is installed in non-editable mode and"
                " we are not able to access it's unit tests. For this command to "
                f"work, please install '{package_name}' in editable mode via: "
                f"'pip install -e <path_to_{package_name}>'",
                file=sys.stderr,
            )
            # We use a print to sys.stderr so monkeypatch unit tests can catch this
            # output
            raise RuntimeError(
                f"Package '{package_name}' isn't installed in editable mode."
            )
        lint_path = str(resources.files("geoips") / "../tests/utils/check_code.sh")
        package_path = str(resources.files(package_name) / "../.")
        for linter in ["bandit", "black", "flake8"]:
            call(["bash", lint_path, linter, package_path], shell=False)


class GeoipsTest(GeoipsCommand):
    """Top-Level Test Command for testing GeoIPS and its corresponding Packages."""

    name = "test"
    command_classes = [GeoipsTestLinting, GeoipsTestScript, GeoipsTestSector]
    # command_classes = [GeoipsTestLinting, GeoipsTestScript, GeoipsTestUnitTest]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utils to handle time stamp information within xarray objects."""

import calendar
from datetime import datetime
from itertools import starmap
import numpy as np
from cftime import num2pydate


def get_posix_from_datetime(dt):
    """Return the POSIX timestamp in seconds.

    Parameters
    ----------
    dt : datetime.datetime
        datetime object to convert to posix timestamp

    Returns
    -------
    long
        representing seconds since 1 January 1970 at 00Z (epoch seconds)
    """
    return calendar.timegm(dt.timetuple())


def get_datetime_from_datetime64(dt64):
    """Get a python datetime object from a numpy datetime64 object.

    Parameters
    ----------
    dt64 : numpy.datetime64'
        numpy.datetime64 object

    Returns
    -------
    datetime.datetime
        Python datetime object

    Notes
    -----
    Backwards compatible with numpy versions
    """
    scale = 1e-9
    if "[ns]" in dt64.dtype.name:
        scale = 1e-9

    return datetime.utcfromtimestamp(dt64.astype(int) * scale)


def get_min_from_xarray_time(xarray_obj, varname):
    """Get the minimum time as a datetime object from xarray object.

    Parameters
    ----------
    xarray_obj : xarray.Dataset or xarray.DataArray
        xarray object from which to extract the minimum time
    varname : str
        Timestamp variable name from which to extract the minimum time

    Returns
    -------
    datetime.datetime
        Python datetime.datetime object representing minimum time of the
        Dataset or DataArray
    """
    # If requested varname is not even in the xarray object, just
    # return None rather than failing.
    if varname not in xarray_obj.variables:
        return None
    minval = xarray_obj[varname].min().to_dict()["data"]
    # Hack to get around bug in most recent version of numpy
    if minval is None:
        goodinds = np.ma.where(xarray_obj[varname].to_masked_array())
        minval = get_datetime_from_datetime64(
            xarray_obj[varname].to_masked_array()[goodinds].min()
        )
    return minval


def get_max_from_xarray_time(xarray_obj, varname):
    """Get the maximum time as a datetime object from xarray object.

    Parameters
    ----------
    xarray_obj : xarray.Dataset or xarray.DataArray
        xarray object from which to extract the maximum time
    varname : str
        Timestamp variable name from which to extract the maximum time

    Returns
    -------
    datetime.datetime
        Python datetime.datetime object representing maximum time of the
        Dataset or DataArray
    """
    # If requested varname is not even in the xarray object, just
    # return None rather than failing.
    if varname not in xarray_obj.variables:
        return None
    maxval = xarray_obj[varname].max().to_dict()["data"]
    # Hack to get around bug in most recent version of numpy
    if maxval is None:
        goodinds = np.ma.where(xarray_obj[varname].to_masked_array())
        maxval = get_datetime_from_datetime64(
            xarray_obj[varname].to_masked_array()[goodinds].max()
        )
    return maxval


def fix_datetime(inxr):
    """Masks the input xarray that has improper datetimes (negatives).

    Parameters
    ----------
    inxr : xarray.Dataset
        Input xarray with datetimes in raw float64 value

    Returns
    -------
    xarray.Dataset
        Xarray with times reformatted as datetimes

    Notes
    -----
    Depreciated versions of xarray cast this as a functional NaT value
    """
    tmp_mask = np.where(inxr.time.values < 0, np.nan, inxr.time.values)
    refactor_date = list(starmap(xarray_to_time, zip(tmp_mask, np.isnan(tmp_mask))))
    inxr.time.values = np.asarray(refactor_date)

    return inxr


def xarray_to_time(inarr, mask):
    """Convert a input array (inarr) with a bool mask (mask) to datetime.

    Notes
    -----
    Uses a nan mask on the array to convert real values.
    """
    dt64_arr = np.full(inarr.shape, np.datetime64("nat"), dtype="datetime64[ns]")
    dt64_arr[~mask] = num2pydate(inarr[~mask], "seconds since 1990-01-01 00:00:00")
    return dt64_arr



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Colormappers interface module."""

from geoips.interfaces.base import BaseModuleInterface


class ColormappersInterface(BaseModuleInterface):
    """Interface for the colormapper to apply to the resulting product."""

    name = "colormappers"
    required_args = {
        "matplotlib": [],
    }

    required_kwargs = {
        "matplotlib": [],
    }

    allowable_kwargs = {
        "matplotlib": [
            "data_range",
            "cmap_name",
            "ascii_path",
            "create_colorbar",
            "cbar_label",
            "cbar_ticks",
            "cbar_tick_labels",
            "cbar_spacing",
            "cbar_full_width",
            "colorbar_kwargs",
            "set_ticks_kwargs",
            "set_label_kwargs",
        ]
    }


colormappers = ColormappersInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `list interfaces` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsListInterfaces(BaseCliTest):
    """Unit Testing Class for List Interfaces Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the GeoipsListInterfaces command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            base_args = ["geoips", "list", "interfaces"]
            alias_args = ["geoips", "ls", "interfaces"]
            self._cmd_list = [base_args, alias_args]
            for argset in [base_args, alias_args]:
                for pkg_name in self.plugin_package_names:
                    self._cmd_list.append(argset + ["-i", "-p", pkg_name])
            # Add argument list which selects certain columns for generic interfaces
            self._cmd_list.append(
                base_args
                + [
                    "--columns",
                    "package",
                    "interface",
                    "supported_famlies",
                    "abspath",
                ]
            )
            self._cmd_list.append(
                alias_args
                + [
                    "--columns",
                    "package",
                    "interface",
                    "supported_famlies",
                    "abspath",
                ]
            )
            # Add argument list which selects certain columns for implemented interfaces
            self._cmd_list.append(
                base_args
                + [
                    "-i",
                    "-p",
                    "geoips",
                    "--columns",
                    "interface",
                    "plugin_type",
                ]
            )
            self._cmd_list.append(
                alias_args
                + [
                    "-i",
                    "-p",
                    "geoips",
                    "--columns",
                    "interface",
                    "plugin_type",
                ]
            )
            self._cmd_list.append(base_args + ["-p", "geoips"])
            self._cmd_list.append(alias_args + ["-p", "geoips"])
            # Add argument list with an invalid command call ("--long" with "--columns")
            self._cmd_list.append(base_args + ["--long", "--columns", "relpath"])
            self._cmd_list.append(alias_args + ["--long", "--columns", "relpath"])
            # Add argument list which invokes the help message for this command
            self._cmd_list.append(base_args + ["-h"])
            self._cmd_list.append(alias_args + ["-h"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips list interfaces ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # bad command has been provided, check the contents of the error message
        assert args != ["geoips", "list", "interfaces"] and args != [
            "geoips",
            "ls",
            "interfaces",
        ]
        assert args != ["geoips", "list", "interfaces", "-i"]
        assert "usage: To use, type `geoips list interfaces`" in error
        if "--long" in args and "--columns" in args:
            assert (
                "error: argument --columns/-c: not allowed with argument --long/-l"
                in error
            )
        elif "-p" in args and "-i" not in args:
            assert "You cannot use the `-p` flag without the `-i` flag." in error

    def check_output(self, args, output):
        """Ensure that the 'geoips list interfaces ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        if "usage: To use, type" in output:
            # -h has been called, check help message contents for this command
            assert "-h" in args
            assert "To use, type `geoips list interfaces`" in output
        else:
            # The args provided are valid, so test that the output is actually correct
            selected_cols = self.retrieve_selected_columns(args)
            if "-i" in args or "-p" in args:
                headers = {
                    "GeoIPS Package": "package",
                    "Interface Type": "plugin_type",
                    "Interface Name": "interface",
                }
            else:
                # `geoips list-interfaces` was called, check for the correct headers
                headers = {
                    "GeoIPS Package": "package",
                    "Interface Type": "plugin_type",
                    "Interface Name": "interface",
                    "Supported Families": "supported_families",
                    "Docstring": "docstring",
                    "Absolute Path": "abspath",
                }
            # Assert that the correct headers exist in the CLI output
            self.assert_correct_headers_in_output(output, headers, selected_cols)


test_sub_cmd = TestGeoipsListInterfaces()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips list interfaces ...' commands.

    This test covers every valid combination of commands for the
    'geoips list interfaces' command. We also test invalid commands, to ensure that
    the proper help documentation  is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'list', 'interfaces'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module for handling optional dependencies throughout GeoIPS."""

import logging
from contextlib import contextmanager
import traceback

LOG = logging.getLogger(__name__)


@contextmanager
def import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and log the event if the import fails.

    Parameters
    ----------
    loglevel : str
        Name of the log level to write to. May be any valid log level (e.g. debug, info,
        etc.).
    """
    try:
        yield None
    except ImportError as err:
        tb = traceback.extract_tb(err.__traceback__)
        filename, lineno, _, _ = tb[-1]
        err_str = f"Failed to import {err.name} at {filename}:{lineno}. "
        err_str += "If you need it, install it."

        getattr(LOG, loglevel)(err_str)
        # print(err_str)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Output formatters interface module."""

from geoips.interfaces.base import BaseModuleInterface


class OutputFormattersInterface(BaseModuleInterface):
    """Data format for the resulting output product (e.g. netCDF, png)."""

    name = "output_formatters"
    required_args = {
        "image": ["area_def", "xarray_obj", "product_name", "output_fnames"],
        "unprojected": ["xarray_obj", "product_name", "output_fnames"],
        "image_overlay": ["area_def", "xarray_obj", "product_name", "output_fnames"],
        "image_multi": [
            "area_def",
            "xarray_obj",
            "product_names",
            "output_fnames",
            "mpl_colors_info",
        ],
        "xrdict_area_varlist_to_outlist": ["xarray_dict", "area_def", "varlist"],
        "xrdict_area_product_outfnames_to_outlist": [
            "xarray_dict",
            "area_def",
            "product_name",
            "output_fnames",
        ],
        "xrdict_area_product_to_outlist": [
            "xarray_dict",
            "area_def",
            "product_name",
        ],
        "xrdict_to_outlist": [
            "xarray_dict",
        ],
        "xrdict_varlist_outfnames_to_outlist": [
            "xarray_dict",
            "varlist",
            "output_fnames",
        ],
        "xarray_data": ["xarray_obj", "product_names", "output_fnames"],
        "standard_metadata": [
            "area_def",
            "xarray_obj",
            "metadata_yaml_filename",
            "product_filename",
        ],
    }
    required_kwargs = {
        "image": ["product_name_title", "mpl_colors_info", "existing_image"],
        "unprojected": ["product_name_title", "mpl_colors_info"],
        "image_overlay": [
            "product_name_title",
            "clean_fname",
            "mpl_colors_info",
            "clean_fname",
            "feature_annotator",
            "gridline_annotator",
            "clean_fname",
            "product_datatype_title",
            "clean_fname",
            "bg_data",
            "bg_mpl_colors_info",
            "clean_fname",
            "bg_xarray",
            "bg_product_name_title",
            "bg_datatype_title",
            "clean_fname",
            "remove_duplicate_minrange",
        ],
        "image_multi": ["product_name_titles"],
        "xarray_dict_data": ["append", "overwrite"],
        "xarray_dict_to_image": [],
        "xarray_data": [],
        "standard_metadata": ["metadata_dir", "basedir", "output_dict"],
        "xrdict_varlist_outfnames_to_outlist": [],
        "xrdict_area_varlist_to_outlist": [],
        "xrdict_area_product_outfnames_to_outlist": [],
        "xrdict_area_product_to_outlist": [],
        "xrdict_to_outlist": [],
    }


output_formatters = OutputFormattersInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Module containing WV (water vapor) algorithm colormap."""

import logging

LOG = logging.getLogger(__name__)

interface = "colormappers"
family = "matplotlib"
name = "WV"


def call(data_range=[-70.0, 0.0]):
    """Colormap developed for displaying algorithms/WV.py processed data.

    Parameters
    ----------
    data_range : list of float, default=[-70, 0]
        * Min and max value for colormap.
        * Ensure the data range matches the range of the algorithm specified for
          use with this colormap
        * This colormap MUST include -70 and 0

    Returns
    -------
    mpl_colors_info : dict
        Dictionary of matplotlib plotting parameters, to ensure consistent
        image output
    """
    min_tb = int(data_range[0])
    max_tb = int(data_range[1])

    if min_tb > -70 or max_tb < 0:
        raise ("Infrared TB range must include -70 and 0")

    from geoips.image_utils.colormap_utils import create_linear_segmented_colormap

    transition_vals = [
        (min_tb, -69.9),
        (-69.9, -50),
        (-50, -40),
        (-40, -35),
        (-35, -30),
        (-30, -20),
        (-20, -10),
        (-10, -0.1),
        (-0.1, max_tb),
    ]
    # LOG.info("inside util= ", max_tb)

    # matching TerraScan Color scheme: noaa_bd_151
    transition_colors = [
        ("#800000", "#800000"),
        ("#820200", "#FFB266"),
        ("#FFB266", "#FFFF7B"),
        ("#FFFF80", "#66FFB2"),
        ("#66FFB2", "#80FFFC"),
        ("#80FFFC", "#3399FF"),
        ("#3399FF", "#146CEB"),
        ("#146CEB", "#7D0382"),
        ("black", "black"),
    ]

    # special selection of label

    ticks = [min_tb, -60, -50, -40, -30, -20, -10, max_tb]

    # selection of min and max values for colormap if needed
    min_tb = transition_vals[0][0]
    max_tb = transition_vals[-1][1]
    ticks = ticks + [int(max_tb)]

    LOG.info("Setting cmap")
    mpl_cmap = create_linear_segmented_colormap(
        "WV_cmap", min_tb, max_tb, transition_vals, transition_colors
    )

    LOG.info("Setting norm")
    from matplotlib.colors import Normalize

    mpl_norm = Normalize(vmin=min_tb, vmax=max_tb)

    cbar_label = r"BT ($^\circ$C)"

    # Must be uniform or proportional, None not valid for Python 3
    cbar_spacing = "proportional"
    mpl_tick_labels = None
    mpl_boundaries = None

    # from geoips.image_utils.mpl_utils import create_colorbar
    # only create colorbar for final imagery
    # cbar = create_colorbar(fig, mpl_cmap, mpl_norm, ticks, cbar_label=cbar_label)
    mpl_colors_info = {
        "cmap": mpl_cmap,
        "norm": mpl_norm,
        "cbar_ticks": ticks,
        "cbar_tick_labels": mpl_tick_labels,
        "cbar_label": cbar_label,
        "boundaries": mpl_boundaries,
        "cbar_spacing": cbar_spacing,
        "colorbar": True,
        "cbar_full_width": True,
    }

    # return cbar, min_tb, max_tb
    return mpl_colors_info
